from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

import numpy as np

from prob_conf_mat.metrics.abc import Metric
from prob_conf_mat.stats import numpy_batched_harmonic_mean


# ==============================================================================
# Fundamental metrics
# Pretty much all metrics are going to rely on these
# ==============================================================================
class DiagMass(Metric):
    r"""Computes the mass on the diagonal of the normalized confusion matrix.

    It is defined as the rate of true positives to all entries:

    $$\mathtt{diag}(\mathbf{CM})=TP / N$$

    where $TP$ are the true positives, and $N$ are the total number of predictions.

    This is a metric primarily used as a intermediate value for other metrics, and says relatively
    little on its own.

    Not to be confused with the True Positive Rate.

    """

    full_name = "Diagonal of Normalized Confusion Matrix"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("norm_confusion_matrix",)
    sklearn_equivalent = None
    aliases = ["diag_mass"]

    def compute_metric(
        self,
        norm_confusion_matrix: jtyping.Float[
            np.ndarray,
            " num_samples num_classes num_classes",
        ],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes"]:
        diag_mass = np.diagonal(
            a=norm_confusion_matrix,
            axis1=1,
            axis2=2,
        )

        return diag_mass


class Prevalence(Metric):
    r"""Computes the marginal distribution of condition occurence. Also known as the prevalence.

    It can be defined as the rate of positives to all predictions:

    $$\mathtt{Prev}=P / N$$

    where $P$ is the count of condition positives, and $N$ are the total number of predictions.

    This is a metric primarily used as a intermediate value for other metrics, and say relatively little
    on its own.

    """  # noqa: E501

    full_name = "Marginal Distribution of Condition"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("p_condition",)
    sklearn_equivalent = None
    aliases = ["prevalence"]

    def compute_metric(
        self,
        p_condition: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes"]:
        return p_condition


class ModelBias(Metric):
    r"""Computes the marginal distribution of prediction occurence. Also known as the model bias.

    It can be defined as the rate of predicted positives to all predictions:

    $$\mathtt{Bias}=PP / N$$

    where $PP$ is the count of predicted positives, and $N$ are the total number of predictions.

    This is a metric primarily used as a intermediate value for other metrics, and say relatively little
    on its own.

    """  # noqa: E501

    full_name = "Marginal Distribution of Predictions"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("p_pred",)
    sklearn_equivalent = None
    aliases = ["model_bias"]

    def compute_metric(
        self,
        p_pred: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes"]:
        # TODO: check confusion matrix before metric computation
        # if (p_pred == 0).any():
        #    warnings.warn("Simulated model neglects class, `p_pred' contains 0.")

        return p_pred


# ==============================================================================
# Simple metrics
# These tell us a little about model performanec, but not the whole story
# Can be computed directly on the fundametal metrics, but are still usually
# used as intermediate values
# ==============================================================================
class TruePositiveRate(Metric):
    r"""Computes the True Positive Rate, also known as recall, sensitivity.

    It is defined as the ratio of correctly predited positives to all condition positives:

    $$\mathtt{TPR}=TP / P$$

    where $TP$ are the true positives, and $TN$ are true negatives and $N$ the number of predictions.

    Essentially, out of all condition positives, how many were correctly predicted. Can be seen as a metric
    measuring retrieval.

    Examples:
        - `tpr`
        - `recall@macro`

    Note: Read more:
        1. [scikit-learn](https://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html)
        2. [Wikipedia](https://en.wikipedia.org/wiki/Precision_and_recall)
    """  # noqa: E501

    full_name = "True Positive Rate"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("p_pred_given_condition",)
    sklearn_equivalent = None
    aliases = ["true_positive_rate", "sensitivity", "recall", "hit_rate", "tpr"]

    def compute_metric(
        self,
        p_pred_given_condition: jtyping.Float[
            np.ndarray,
            " num_samples num_classes num_classes",
        ],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes"]:
        true_positive_rate = np.diagonal(
            a=p_pred_given_condition,
            axis1=1,
            axis2=2,
        )

        return true_positive_rate


class FalseNegativeRate(Metric):
    r"""Computes the False Negative Rate, also known as the miss-rate.

    It is defined as the ratio of false negatives to condition positives:

    $$\mathtt{FNR}=FN / (TP + FN)$$

    where $TP$ are the true positives, and $FN$ are the false negatives.

    Examples:
        - `fnr`
        - `false_negative_rate@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/False_positives_and_false_negatives#Related_terms)
    """  # noqa: E501

    full_name = "False Negative Rate"
    is_complex = True
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("true_positive_rate",)
    sklearn_equivalent = None
    aliases = ["false_negative_rate", "miss_rate", "fnr"]

    def compute_metric(
        self,
        true_positive_rate: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        false_negative_rate = 1 - true_positive_rate

        return false_negative_rate


class PositivePredictiveValue(Metric):
    r"""Computes the Positive Predictive Value, also known as precision.

    It is defined as the ratio of true positives to predicted positives:

    $$\mathtt{PPV}=TP / (TP + FP)$$

    where $TP$ is the count of true positives, and $FP$ the count falsely predicted positives.

    It is the complement of the False Discovery Rate, $PPV=1-FDR$.

    Examples:
        - `ppv`
        - `precision@macro`

    Note: Read more:
        1. [scikit-learn](https://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html)
        2. [Wikipedia](https://en.wikipedia.org/wiki/Precision_and_recall)
    """  # noqa: E501

    full_name = "Positive Predictive Value"
    is_complex = True
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("p_condition_given_pred",)
    sklearn_equivalent = None
    aliases = ["positive_predictive_value", "precision", "ppv"]

    def compute_metric(
        self,
        p_condition_given_pred: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        positive_predictive_value = np.diagonal(
            a=p_condition_given_pred,
            axis1=1,
            axis2=2,
        )

        return positive_predictive_value


class FalseDiscoveryRate(Metric):
    r"""Computes the False Discovery Rate.

    It is defined as the ratio of falsely predicted positives to predicted positives:

    $$\mathtt{FDR}=FP / (TP + FP)$$

    where $TP$ is the count of true positives, and $FP$ the count of falsely predicted positives.

    It is the complement of the Positive Predictve Value, $FDR=1-PPV$.

    Examples:
        - `fdr`
        - `false_discovery_rate@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/False_discovery_rate)
    """  # noqa: E501

    full_name = "False Discovery Rate"
    is_complex = True
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("positive_predictive_value",)
    sklearn_equivalent = None
    aliases = ["false_discovery_rate", "fdr"]

    def compute_metric(
        self,
        positive_predictive_value: jtyping.Float[
            np.ndarray,
            " num_samples num_classes",
        ],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        false_discovery_rate = 1 - positive_predictive_value

        return false_discovery_rate


class FalsePositiveRate(Metric):
    r"""Computes the False Positive Rate, the probability of false alarm.

    Also known as the fall-out.

    It is defined as the ratio of falsely predicted positives to condition negatives:

    $$\mathtt{FPR}=FP / (TN + FP)$$

    where $TN$ is the count of true negatives, and $FP$ the count of falsely predicted positives.

    It is the complement of the True Negative Rate, $FPR=1-TNR$.

    Examples:
        - `fpr`
        - `fall-out@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/False_positive_rate)
    """  # noqa: E501

    full_name = "False Positive Rate"
    is_complex = True
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("diag_mass", "p_pred", "p_condition")
    sklearn_equivalent = None
    aliases = ["false_positive_rate", "fall-out", "fall_out", "fpr"]

    def compute_metric(
        self,
        diag_mass: jtyping.Float[np.ndarray, " num_samples num_classes"],
        p_pred: jtyping.Float[np.ndarray, " num_samples num_classes"],
        p_condition: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        false_positive_rate = (p_pred - diag_mass) / (1 - p_condition)

        return false_positive_rate


class TrueNegativeRate(Metric):
    r"""Computes the True Negative Rate, i.e. specificity, selectivity.

    It is defined as the ratio of true predicted negatives to condition negatives:

    $$\mathtt{TNR}=TN / (TN + FP)$$

    where $TN$ is the count of true negatives, and FP the count of falsely predicted positives.

    It is the complement of the False Positive Rate, $TNR=1-FPR$.

    Examples:
        - `tnr`
        - `selectivity@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Sensitivity_and_specificity)
    """  # noqa: E501

    full_name = "True Negative Rate"
    is_complex = True
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("false_positive_rate",)
    sklearn_equivalent = None
    aliases = ["true_negative_rate", "specificity", "selectivity", "tnr"]

    def compute_metric(
        self,
        false_positive_rate: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        true_negative_rate = 1 - false_positive_rate

        return true_negative_rate


class FalseOmissionRate(Metric):
    r"""Computes the False Omission Rate.

    It is defined as the ratio of falsely predicted negatives to all predicted negatives:

    $$\mathtt{FOR}=FN / (TN + FN)$$

    where $$TN$$ is the count of true negatives, and $$FN$$ the count of falsely predicted negatives.

    It is the complement of the Negative Predictive Value, $FOR=1-NPV$.

    Examples:
        - `for`
        - `false_omission_rate@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Positive_and_negative_predictive_values#false_omission_rate)
    """  # noqa: E501

    full_name = "False Omission Rate"
    is_complex = True
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("p_condition", "p_pred", "diag_mass")
    sklearn_equivalent = None
    aliases = ["false_omission_rate", "for"]

    def compute_metric(
        self,
        p_condition: jtyping.Float[np.ndarray, " num_samples num_classes"],
        p_pred: jtyping.Float[np.ndarray, " num_samples num_classes"],
        diag_mass: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        # This requires reasoning about true negatives in a multi-class setting
        # This is somewhat involved, hence the unintuitive formula
        false_omission_rate = (p_condition - diag_mass) / (1 - p_pred)

        return false_omission_rate


class NegativePredictiveValue(Metric):
    r"""Computes the Negative Predicitive Value.

    It is defined as the ratio of true negatives to all predicted negatives:

    $$\mathtt{NPV}=TN / (TN + FN)$$

    where TN are the true negatives, and FN are the falsely predicted negatives.

    It is the complement of the False Omission Rate, $NPV=1-FOR$.

    Examples:
        - `npv`
        - `negative_predictive_value@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Positive_and_negative_predictive_values#Negative_predictive_value_(NPV))
    """  # noqa: E501

    full_name = "Negative Predictive Value"
    is_complex = True
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("false_omission_rate",)
    sklearn_equivalent = None
    aliases = ["negative_predictive_value", "npv"]

    def compute_metric(
        self,
        false_omission_rate: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        # This requires reasoning about true negatives in a multi-class setting
        # This is somewhat involved, hence the unintuitive formula
        negative_predictive_value = 1 - false_omission_rate

        return negative_predictive_value


# ==============================================================================
# Complex metrics
# These actually tell you something interesting about model performance
# ==============================================================================
class Accuracy(Metric):
    r"""Computes the multiclass accuracy score.

    It is defined as the rate of correct classifications to all classifications:

    $$\mathtt{Acc}=(TP + TN) / N$$

    where $TP$ are the true positives, $TN$ the true negatives and $N$ the total number of predictions.

    Possible values lie in the range [0.0, 1.0], with larger values denoting better performance. The value
    of a random classifier is dependent on the label distribution, which makes accuracy especially susceptible
    to class imbalance. It is also not directly comparable across datasets.

    Examples:
        - `acc`
        - `accuracy@macro`

    Note: Read more:
        1. [scikit-learn](https://scikit-learn.org/stable/modules/model_evaluation.html#accuracy-score)
        2. [Wikipedia](https://en.wikipedia.org/wiki/Accuracy_and_precision#In_binary_classification)
    """  # noqa: E501

    full_name = "Accuracy"
    is_multiclass = True
    bounds = (0.0, 1.0)
    dependencies = ("diag_mass",)
    sklearn_equivalent = "accuracy_score"
    aliases = ["acc", "accuracy"]

    def compute_metric(
        self,
        diag_mass: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples 1"]:
        acc = np.sum(diag_mass, axis=1)
        return acc[:, np.newaxis]


class BalancedAccuracy(Metric):
    r"""Computes the balanced accuracy score.

    It is defined as the the arithmetic average of the per-class true-positive rate:

    $$\mathtt{BA}=\frac{1}{|C|}\sum TPR_{c}$$

    where $TPR$ is the true positive rate (precision).

    Possible values lie in the range [0.0, 1.0], with larger values denoting better performance. Unlike
    accuracy, balanced accuracy can be 'chance corrected', such that random performance is yield a score
    of 0.0. This can be achieved by setting `adjusted=True`.

    Examples:
        - `ba`
        - `balanced_accuracy@macro`
        - `ba+adjusted=True`

    Note: Read more:
        1. [scikit-learn](https://scikit-learn.org/stable/modules/model_evaluation.html#balanced-accuracy-score)

    Args:
        adjusted (bool): whether the chance-corrected variant is computed. Defaults to `False`.
    """  # noqa: E501

    full_name = "Balanced Accuracy"
    is_multiclass = True
    bounds = (0.0, 1.0)
    dependencies = ("tpr", "p_condition")
    sklearn_equivalent = "balanced_accuracy_score"
    aliases = ["ba", "balanced_accuracy"]

    def __init__(self, *, adjusted: bool = False) -> None:
        super().__init__()
        self.adjusted = adjusted

    def _compute_ba(
        self,
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples 1"]:
        balanced_accuracy = np.nanmean(
            tpr,
            axis=-1,
        )
        return balanced_accuracy

    def compute_metric(
        self,
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        p_condition: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples 1"]:
        ba = self._compute_ba(tpr)
        if self.adjusted:
            chance = 1 / (p_condition != 0).sum(axis=1)
            ba = (ba - chance) / (1 - chance)

        return ba[:, np.newaxis]


class MatthewsCorrelationCoefficient(Metric):
    """Computes the multiclass Matthew's Correlation Coefficient (MCC), also known as the phi coefficient.

    Goes by a variety of names, depending on the application scenario.

    A metric that holistically combines many different classification metrics.

    A perfect classifier scores 1.0, a random classifier 0.0. Values smaller than 0
    indicate worse than random performance.

    It's absolute value is proportional to the square root of the Chi-square test statistic.

    Quoting Wikipedia:
    > Some scientists claim the Matthews correlation coefficient to be the most informative
    single score to establish the quality of a binary classifier prediction in a confusion matrix context.

    Examples:
        - `mcc`
        - `phi`

    Note: Read more:
        1. [scikit-learn](https://scikit-learn.org/stable/modules/model_evaluation.html#matthews-correlation-coefficient)
        2. [Wikipedia](https://en.wikipedia.org/wiki/Phi_coefficient)

    """  # noqa: E501

    full_name = "Matthews Correlation Coefficient"
    is_multiclass = True
    bounds = (-1.0, 1.0)
    dependencies = ("diag_mass", "p_condition", "p_pred")
    sklearn_equivalent = "matthews_corrcoef"
    aliases = [
        "mcc",
        "matthews_corrcoef",
        "matthews_correlation_coefficient",
        "phi",
        "phi_coefficient",
    ]

    def compute_metric(
        self,
        diag_mass: jtyping.Float[np.ndarray, "num_samples num_classes"],
        p_condition: jtyping.Float[np.ndarray, "num_samples num_classes"],
        p_pred: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples 1"]:
        marginals_inner_prod = np.einsum("bc, bc->b", p_condition, p_pred)
        numerator = np.sum(diag_mass, axis=1) - marginals_inner_prod

        mcc = numerator / np.sqrt(
            (1 - np.power(p_condition, 2).sum(axis=1))
            * (1 - np.power(p_pred, 2).sum(axis=1)),
        )

        return mcc[:, np.newaxis]


class CohensKappa(Metric):
    r"""Computes the multiclass Cohen's Kappa coefficient.

    Commonly used to quantify inter-annotator agreement, Cohen's kappa can also
    be used to quantify the quality of a predictor.

    It is defined as

    $$\kappa=\frac{p_o-p_e}{1-p_e}$$

    where $p_o$ is the observed agreement and $p_e$ the expected agreement
    due to chance. Perfect agreement yields a score of 1, with a score of
    0 corresponding to random performance. Several guidelines exist to interpret
    the magnitude of the score.

    Examples:
        - `kappa`
        - `cohen_kappa`

    Note: Read more:
        1. [sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html#cohen-kappa)
        2. [Wikipedia](https://en.wikipedia.org/wiki/Cohen%27s_kappa)
    """

    full_name = "Cohen's Kappa"
    is_multiclass = True
    bounds = (-1.0, 1.0)
    dependencies = ("diag_mass", "p_condition", "p_pred")
    sklearn_equivalent = "cohen_kappa_score"
    aliases = ["kappa", "cohen_kappa"]

    def compute_metric(
        self,
        diag_mass: jtyping.Float[np.ndarray, "num_samples num_classes"],
        p_condition: jtyping.Float[np.ndarray, "num_samples num_classes"],
        p_pred: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples 1"]:
        p_agreement = np.sum(diag_mass, axis=1)

        p_chance = np.einsum("bc, bc->b", p_condition, p_pred)

        kappa = (p_agreement - p_chance) / (1 - p_chance)

        return kappa[:, np.newaxis]


class F1(Metric):
    r"""Computes the univariate $F_{1}$-score.

    It is defined as:

    $$\mathtt{F}_{1}=2\dfrac{\text{precision} \cdot \text{recall}}{\text{precision} + \text{recall}}$$

    or simply put, the harmonic mean between precision (PPV) and recall (TPR).

    It is an exceedingly common metric used to evaluate machine learning performance. It is closely
    related to the Precision-Recall curve, an anlysis with varying thresholds.

    The 1 in the name from an unseen $\beta$ parameter that weights precision and recall.
    See the `FBeta` metric.

    The $F_{1}$-score is susceptible to class imbalance. Values fall in the range [0, 1]. A random
    classifier which predicts a class with a probability $p$, achieves a performance of,

    $$2\dfrac{\text{prevalence}\cdot p}{\text{prevalence}+p}.$$

    Since this value is maximized for $p=1$, [Flach & Kull](https://proceedings.neurips.cc/paper/2015/hash/33e8075e9970de0cfea955afd4644bb2-Abstract.html)
    recommend comparing performance not to a random classifier, but the 'always-on' classifier
    (perfect recall but poor precision). See the `F1Gain` metric.

    Examples:
        - `f1`
        - `f1@macro`

    Note: Read more:
        1. [sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html#precision-recall-f-measure-metrics)
        2. [Wikipedia](https://en.wikipedia.org/wiki/F-score)
    """

    full_name = "F1-score"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("ppv", "tpr")
    sklearn_equivalent = "f1_score"
    aliases = ["f1"]

    def compute_metric(
        self,
        ppv: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        f1 = 2 * (ppv * tpr) / (ppv + tpr)

        # In case one of the ratios is nan (most likely due to 0 division), set to 0
        f1 = np.nan_to_num(
            f1,
            nan=0.0,
        )

        return f1


class FBeta(Metric):
    r"""Computes the univariate $F_{\beta}$-score.

    Commonly used to quantify inter-annotator agreement, Cohen's kappa can also
    be used to quantify the quality of a predictor.

    It is defined as:

    $$\mathtt{F}_{\beta}=(1+\beta^2)\dfrac{\text{precision} \cdot \text{recall}}{\beta^2\cdot\text{precision} + \text{recall}}$$

    or simply put, the weighted harmonic mean between precision (PPV) and recall (TPR).

    The value of $\beta$ determines to which degree a user deems recall more important than
    precision. Larger values (x > 1) weight recall more, whereas lower values weight precision more.
    A value of 1 corresponds to equal weighting, see the `F1` metric.

    The $F_{\beta}$-score is susceptible to class imbalance. Values fall in the range [0, 1]. A
    random classifier which predicts a class with a probability $p$, achieves a performance of,

    $$(1+\beta^2)\dfrac{\text{prevalence}\cdot p}{\beta^2\cdot\text{prevalence}+p}.$$

    Since this value is maximized for $p=1$, [Flach & Kull](https://proceedings.neurips.cc/paper/2015/hash/33e8075e9970de0cfea955afd4644bb2-Abstract.html)
    recommend comparing performance not to a random classifier, but the 'always-on' classifier
    (perfect recall but poor precision). See the `FBetaGain` metric.

    Examples:
        - `fbeta+beta=2`
        - `fbeta+beta=0.5@macro`

    Note: Read more:
        1. [sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html#precision-recall-f-measure-metrics)
        2. [Wikipedia](https://en.wikipedia.org/wiki/F-score)
    """  # noqa: E501

    full_name = "FBeta-score"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("ppv", "tpr")
    sklearn_equivalent = "fbeta_score"
    aliases = ["fbeta"]

    def __init__(self, beta: float = 1.0):
        super().__init__()

        self.beta = beta

    def compute_metric(
        self,
        ppv: jtyping.Float[np.ndarray, " num_samplesnum_classes"],
        tpr: jtyping.Float[np.ndarray, " num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        beta_2 = self.beta**2

        fbeta = (1 + beta_2) * (ppv * tpr) / (beta_2 * ppv + tpr)

        # In case one of the ratios is nan (most likely due to 0 division), set to 0
        fbeta = np.nan_to_num(
            fbeta,
            nan=0.0,
        )

        return fbeta


class Informedness(Metric):
    r"""Computes the Informedness metric, also known Youden's J.

    It is defined as:

    $$\mathtt{J}=\text{sensitivity}+\text{specificity}-1$$

    where sensitivity is the True Positive Rate (TPR), and specificity is the
    True Negative Rate (TNR).

    Values fall in the range [-1, 1], with higher values corresponding to better performance and 0
    corresponding to random performance.

    In the binary case, this metric is equivalent to the adjusted balanced accuracy, `ba+adj=True`.

    It is commonly used in conjunction with a Reciever-Operator Curve analysis.

    Examples:
        - `informedness`
        - `youdenj@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Youden%27s_J_statistic)
    """

    full_name = "Informedness"
    is_multiclass = False
    bounds = (-1.0, 1.0)
    dependencies = ("tpr", "tnr")
    sklearn_equivalent = None
    aliases = ["informedness", "youdenj", "youden_j", "bookmaker_informedness", "bm"]

    def compute_metric(
        self,
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tnr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        return tpr + tnr - 1


class Markedness(Metric):
    r"""Computes the markedness metric, also known as $\Delta p$.

    It is defined as:

    $$\Delta p=\text{precision}+NPV-1$$

    where precision is the Positive Predictive Value (PPV).

    Values fall in the range [-1, 1], with higher values corresponding to better performance and 0
    corresponding to random performance.

    It is commonly used in conjunction with a Reciever-Operator Curve analysis.

    Examples:
        - `markedness`
        - `delta_p@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Markedness#)
    """

    full_name = "Markedness"
    is_multiclass = False
    bounds = (-1.0, 1.0)
    dependencies = ("ppv", "npv")
    sklearn_equivalent = None
    aliases = ["markedness", "delta_p"]

    def compute_metric(
        self,
        ppv: jtyping.Float[np.ndarray, "num_samples num_classes"],
        npv: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        return ppv + npv - 1


class P4(Metric):
    r"""Computes the P4 metric.

    It is defined as:

    $$\mathtt{P4}=4\left(\dfrac{1}{\text{precision}}+\dfrac{1}{\text{recall}}+\dfrac{1}{\text{specificity}}+\dfrac{1}{NPV}\right)^{-1}$$

    where precision corresponds to the Positive Predictive Value (PPV), recall to the
    True Positive Rate (TPR), and specificity to the True Negative Rate (TNR). Put otherwise, it is
    the harmonic mean of the 4 listed metrics.

    Introduced in 2022 by [Sitarz](https://arxiv.org/abs/2210.11997), it is meant to extend the
    properties of the F1, Markedness and Informedness metrics. It is one of few defined metrics
    that incorporates the Negative Predictive Value.

    Possible values lie in the range [0, 1], with a score of 0 implying one of the intermediate
    metrics is 0, and a 1 requiring perfect classification.

    Relative to MCC, the author notes different behaviour at extreme values, but otherwise the
    metrics are meant to provide a similar amount of information with a single value.

    Examples:
        - `p4`
        - `p4@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/P4-metric)

    """

    full_name = "P4-score"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("ppv", "tpr", "tnr", "npv")
    sklearn_equivalent = None
    aliases = ["p4"]

    def compute_metric(
        self,
        ppv: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tnr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        npv: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        values = np.stack(
            [
                ppv,
                tpr,
                tnr,
                npv,
            ],
            axis=2,
        )

        return numpy_batched_harmonic_mean(values, axis=2, keepdims=False)


class JaccardIndex(Metric):
    r"""Computes the Jaccard Index, also known as the threat score.

    It is defined as:

    $$\mathtt{Jaccard}=\dfrac{TP}{TP+FP+FN}$$

    where $TP$ is the count of true positives, $FP$ the count of false positives and $FN$ the count
    of false negatives.

    Alternatively, it may be defined as the area of overlap between predicted and conditions,
    divided by the area of all predicted and condition positives.

    Due to the alternative definition, it is commonly used when labels are not readily present, for
    example in evaluating clustering performance.

    Examples:
        - `jaccard`
        - `critical_success_index@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Jaccard_index#Jaccard_index_in_binary_classification_confusion_matrices)

    """

    full_name = "Jaccard Index"
    is_multiclass = False
    bounds = (0.0, 1.0)
    dependencies = ("diag_mass", "p_pred", "p_condition")
    sklearn_equivalent = "jaccard_score"
    aliases = ["jaccard", "jaccard_index", "threat_score", "critical_success_index"]

    def compute_metric(
        self,
        diag_mass: jtyping.Float[np.ndarray, "num_samples num_classes"],
        p_pred: jtyping.Float[np.ndarray, "num_samples num_classes"],
        p_condition: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        return diag_mass / (p_pred + p_condition - diag_mass)


class PositiveLikelihoodRatio(Metric):
    r"""Computes the positive likelihood ratio.

    It is defined as

    $$\mathtt{LR}^{+}=\dfrac{\text{sensitivity}}{1-\text{specificity}}$$

    where sensitivity is the True Positive Rate (TPR), and specificity is the
    True Negative Rate (TNR).

    Simply put, it is the ratio of the probabilities of the model predicting a positive when the
    condition is positive and negative, respectively.

    Possible values lie in the range [0.0, $\infty$], with 0.0 corresponding to no true positives,
    and infinity corresponding to no false positives. Larger values indicate better performance,
    with a score of 1 corresponding to random performance.

    Examples:
        - `plr`
        - `positive_likelihood_ratio@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Likelihood_ratios_in_diagnostic_testing#positive_likelihood_ratio)

    """

    full_name = "Positive Likelihood Ratio"
    is_multiclass = False
    bounds = (0.0, float("inf"))
    dependencies = ("tpr", "fpr")
    sklearn_equivalent = "class_likelihood_ratios"
    aliases = ["plr", "positive_likelihood_ratio"]

    def __init__(self, *, clamp: bool = False) -> None:
        super().__init__()
        self.clamp = clamp

    def compute_metric(
        self,
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        fpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        if self.clamp:
            fpr = np.where(fpr == 0, np.min(fpr[fpr != 0.0]), fpr)

        return tpr / fpr


class LogPositiveLikelihoodRatio(Metric):
    r"""Computes the positive likelihood ratio.

    It is defined as

    $$\mathtt{LogLR}{+}=\log\dfrac{\text{sensitivity}}{1-\text{specificity}}$$

    where sensitivity is the True Positive Rate (TPR), and specificity is the
    True Negative Rate (TNR).

    Simply put, it is logarithm of the ratio of the probabilities of the model predicting a
    positive when the condition is positive and negative, respectively.

    Possible values lie in the range ($-\infty$, $\infty$), with $-\infty$ corresponding to no
    true positives, and infinity corresponding to no false positives. Larger values indicate better
    performance, with a score of 0 corresponding to random performance.

    Examples:
        - `log_plr`
        - `lplr`
        - `log_positive_likelihood_ratio@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Likelihood_ratios_in_diagnostic_testing#positive_likelihood_ratio)

    """

    full_name = "Log Positive Likelihood Ratio"
    is_multiclass = False
    bounds = (-float("inf"), float("inf"))
    dependencies = ("tpr", "fpr")
    sklearn_equivalent = "class_likelihood_ratios"
    aliases = ["log_plr", "lplr", "log_positive_likelihood_ratio"]

    def __init__(self, *, clamp: bool = False) -> None:
        super().__init__()
        self.clamp = clamp

    def compute_metric(
        self,
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        fpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        if self.clamp:
            fpr = np.where(fpr == 0, np.min(fpr[fpr != 0.0]), fpr)

        lplr = np.log(tpr) - np.log(fpr)

        return lplr


class NegativeLikelihoodRatio(Metric):
    r"""Computes the negative likelihood ratio.

    It is defined as

    $$\mathtt{LR}^{-}=\dfrac{1-\text{sensitivity}}{\text{specificity}}$$

    where sensitivity is the True Positive Rate (TPR), and specificity is the
    True Negative Rate(TNR).

    Simply put, it is the ratio of the probabilities of the model predicting a negative when the
    condition is positive and negative, respectively.

    Possible values lie in the range [0.0, $\infty$], with 0.0 corresponding to no false negatives,
    and infinity corresponding to no true negatives. Smaller values indicate better performance,
    with a score of 1 corresponding to random performance.

    Examples:
        - `nlr`
        - `negative_likelihood_ratio@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Likelihood_ratios_in_diagnostic_testing#negative_likelihood_ratio)

    """

    full_name = "Negative Likelihood Ratio"
    is_multiclass = False
    bounds = (0.0, float("inf"))
    dependencies = ("fnr", "tnr")
    sklearn_equivalent = "class_likelihood_ratios"
    aliases = ["negative_likelihood_ratio", "nlr"]

    def __init__(self, *, clamp: bool = False) -> None:
        super().__init__()
        self.clamp = clamp

    def compute_metric(
        self,
        fnr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tnr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        if self.clamp:
            tnr = np.where(tnr == 0, np.min(tnr[tnr != 0.0]), tnr)

        return fnr / tnr


class LogNegativeLikelihoodRatio(Metric):
    r"""Computes the negative likelihood ratio.

    It is defined as

    $$\mathtt{LogLR}{-}=\log \dfrac{1-\text{sensitivity}}{\text{specificity}}$$

    where sensitivity is the True Positive Rate (TPR), and specificity is the
    True Negative Rate (TNR).

    Simply put, it is the logarithm of the ratio of the probabilities of the model predicting a
    negative when the condition is positive and negative, respectively.

    Possible values lie in the range ($-\infty$, $\infty$), with $-\infty$ corresponding to no
    true positives, and infinity corresponding to no true negatives. Smaller values indicate better
    performance, with a score of 0 corresponding to random performance.

    Examples:
        - `log_nlr`
        - `lnlr`
        - `log_negative_likelihood_ratio@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Likelihood_ratios_in_diagnostic_testing#negative_likelihood_ratio)

    """

    full_name = "Log Negative Likelihood Ratio"
    is_multiclass = False
    bounds = (-float("inf"), float("inf"))
    dependencies = ("fnr", "tnr")
    sklearn_equivalent = "class_likelihood_ratios"
    aliases = ["lnlr", "log_negative_likelihood_ratio", "log_nlr"]

    def __init__(self, *, clamp: bool = False) -> None:
        super().__init__()
        self.clamp = clamp

    def compute_metric(
        self,
        fnr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tnr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        if self.clamp:
            tnr = np.where(tnr == 0, np.min(tnr[tnr != 0.0]), tnr)

        lnlr = np.log(fnr) - np.log(tnr)

        return lnlr


class DiagnosticOddsRatio(Metric):
    r"""Computes the diagnostic odds ratio.

    It is defined as:

    $$\mathtt{DOR}=\dfrac{\mathtt{LR}^{+}=}{\mathtt{LR}^{-}=}$$

    where $\mathtt{LR}^{+}=$ and $\mathtt{LR}^{-}=$ are the positive and
    negative likelihood ratios, respectively.

    Possible values lie in the range [0.0, $\infty$]. Larger values indicate better performance,
    with a score of 1 corresponding to random performance.

    To make experiment aggregation easier, you can log transform this metric by specifying
    `log_transform=true`. This makes the sampling distribution essentially Gaussian.

    Examples:
        - `dor`
        - `diagnostic_odds_ratio@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Diagnostic_odds_ratio)

    """

    full_name = "Diagnostic Odds Ratio"
    is_multiclass = False
    bounds = (0.0, float("inf"))
    dependencies = ("nlr", "plr")
    sklearn_equivalent = None
    aliases = ["dor", "diagnostic_odds_ratio"]

    def compute_metric(
        self,
        plr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        nlr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        return plr / nlr


class LogDiagnosticOddsRatio(Metric):
    r"""Computes the diagnostic odds ratio.

    It is defined as:

    $$\mathtt{LogDOR}=\mathtt{LogLR}^{+}-\mathtt{LogLR}^{-}$$

    where $\mathtt{LR}^{+}$ and $\mathtt{LR}^{-}=$ are the positive and
    negative likelihood ratios, respectively.

    Possible values lie in the range (-$\infty$, $\infty$). Larger values indicate better
    performance, with a score of 0 corresponding to random performance.

    Examples:
        - `log_dor`
        - `ldor`
        - `log_diagnostic_odds_ratio@macro`

    Note: Read more:
        1. [Wikipedia](https://en.wikipedia.org/wiki/Diagnostic_odds_ratio)

    """

    full_name = "Log Diagnostic Odds Ratio"
    is_multiclass = False
    bounds = (-float("inf"), float("inf"))
    dependencies = ("log_plr", "log_nlr")
    sklearn_equivalent = None
    aliases = ["log_dor", "ldor", "log_diagnostic_odds_ratio"]

    def compute_metric(
        self,
        log_plr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        log_nlr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        log_dor = log_plr - log_nlr

        return log_dor


class PrevalenceThreshold(Metric):
    r"""Computes the prevalence threshold.

    It is defined as:

    $$\phi \mathtt{e}=\frac{\sqrt{\mathtt{TPR}\cdot(1-\mathtt{TNR})}+\mathtt{TNR}-1}{\mathtt{TPR}+\mathtt{TNR}-1}$$

    where $\mathtt{TPR}$ and $\mathtt{TNR}$ are the true positive and negative rates, respectively.

    Possible values lie in the range (0, 1). Larger values indicate *worse* performance, with a
    score of 0 corresponding to perfect classification, and a score of 1 to perfect
    misclassifcation.

    It representents the inflection point in a sensitivity and specificity curve (ROC), beyond which
    a classifiers positive predictive value drops sharply. See [Balayla (2020)](https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0240215#sec002)
    for more information.

    Examples:
        - `pt`
        - `prevalence_threshold`

    Note: Read more:
        1. [Balayla, J. (2020). Prevalence threshold ($\phi \mathtt{e}$) and the geometry of screening curves. Plos one, 15(10), e0240215.](https://pmc.ncbi.nlm.nih.gov/articles/PMC7540853/)

    """  # noqa: E501

    full_name = "Prevalence Threshold"
    is_multiclass = False
    bounds = (0, 1)
    dependencies = ("tpr", "tnr")
    sklearn_equivalent = None
    aliases = ["prev_thresh", "pt", "prevalence_threshold"]

    def compute_metric(
        self,
        tpr: jtyping.Float[np.ndarray, "num_samples num_classes"],
        tnr: jtyping.Float[np.ndarray, "num_samples num_classes"],
    ) -> jtyping.Float[np.ndarray, " num_samples num_classes num_classes"]:
        numerator = np.sqrt(tpr * (1 - tnr)) + tnr - 1
        denominator = tpr + tnr - 1

        pt = numerator / denominator

        return pt



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

import numpy as np


def histogram_mode_estimator(
    samples: jtyping.Float[np.ndarray, " num_samples"],
    bounds: tuple[float, float] | None = None,
) -> float:
    """"Tries to estimate the mode of a distribution from its samples."""
    bin_counts, bin_edges = np.histogram(samples, bins="auto", range=bounds)
    modal_bin = np.argmax(bin_counts)

    mode = (bin_edges[modal_bin] + bin_edges[modal_bin + 1]) / 2

    return mode


# TODO: remove or integrate KDE mode estimation
# def kde_mode_estimator(
#     samples: jtyping.Float[np.ndarray, " num_samples"],
#     range: typing.Optional[typing.Tuple[float, float]] = None,
# )-> float:
#     from lightkde import kde_1d

#     kernel_densities, evaluation_points = kde_1d(
#         samples,
#         x_min=range[0] if range is not None else None,
#         x_max=range[1] if range is not None else None,
#     )

#     modal_point = np.argmax(kernel_densities)

#     mode = evaluation_points[modal_point]

#     return mode



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

import numpy as np
import scipy.stats as stats
from dataclasses import dataclass
import typing

from prob_conf_mat.stats.mode_estimation import histogram_mode_estimator
from prob_conf_mat.stats.hdi_estimation import hdi_estimator


@dataclass(frozen=True)
class PosteriorSummary:
    """Summary statistics of some probability distribution."""

    median: float
    mode: float
    ci_probability: float
    hdi: tuple[float, float]
    skew: float
    kurtosis: float

    @property
    def metric_uncertainty(self) -> float:
        """The metric uncertainty (MU), defined as the size of the HDI.

        Returns:
            float: the MU
        """
        return self.hdi[1] - self.hdi[0]

    @property
    def headers(self) -> list[str]:
        """The column headers."""
        return [
            "Median",
            "Mode",
            f"{self.ci_probability * 100:.1f}% HDI",
            "MU",
            "Skew",
            "Kurt",
        ]

    def as_dict(self) -> dict[str, float | tuple[float, float]]:
        """Returns the dict representation of the statistics.

        Useful for coverting to a table.

        Returns:
            dict[str, float | tuple[float, float]]
        """
        d = {
            "Median": self.median,
            "Mode": self.mode,
            f"{self.ci_probability * 100:.1f}% HDI": self.hdi,
            "MU": self.metric_uncertainty,
            "Skew": self.skew,
            "Kurt": self.kurtosis,
        }

        return d


def summarize_posterior(
    posterior_samples: jtyping.Float[np.ndarray, " num_samples"],
    ci_probability: float,
) -> PosteriorSummary:
    """Summarizes a distribution, assumed to be a posterior, based on samples from it.

    Args:
        posterior_samples (jtyping.Float[np.ndarray, " num_samples"]): samples from the posterior.
        ci_probability (float): the probability under the HDI.

    Returns:
        PosteriorSummary: the summary statistics.
    """
    summary = PosteriorSummary(
        median=typing.cast("float", np.median(posterior_samples)),
        mode=typing.cast("float", histogram_mode_estimator(samples=posterior_samples)),
        ci_probability=ci_probability,
        hdi=typing.cast(
            "tuple[float, float]",
            hdi_estimator(posterior_samples, prob=ci_probability),
        ),
        skew=stats.skew(posterior_samples),
        kurtosis=stats.kurtosis(posterior_samples),
    )

    return summary



# Taken from https://blog.iqmo.com/blog/python/jupyter_notebook_testing/
import os
from pathlib import Path

import pytest
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor

NOTEBOOK_DIR = Path("./documentation/Getting Started")
SKIP_NOTEBOOKS = []
TIMEOUT = 600

assert NOTEBOOK_DIR.is_dir(), NOTEBOOK_DIR


@pytest.mark.parametrize(
    argnames="notebook",
    argvalues=[
        file
        for file in NOTEBOOK_DIR.resolve().glob("*.ipynb")
        if file.name not in SKIP_NOTEBOOKS
    ],
)
def test_notebook_execution(notebook: Path) -> None:
    # First set the absolute path of the notebook
    notebook_abs_fp = notebook.resolve()

    # Change the working directory to that of the notebook
    os.chdir(notebook.parent)

    with notebook_abs_fp.open() as f:
        nb = nbformat.read(fp=f, as_version=4)

    ep = ExecutePreprocessor(timeout=TIMEOUT)
    ep.preprocess(nb)



from __future__ import annotations
import typing
import warnings

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

    from prob_conf_mat.utils import RNG

import numpy as np
import scipy
import scipy.stats

from prob_conf_mat.experiment_aggregation.abc import ExperimentAggregator
from prob_conf_mat.experiment_aggregation.heterogeneity import (
    heterogeneity_dl,
    heterogeneity_pm,
)
from prob_conf_mat.stats import truncated_sample

__all__ = [
    "SingletonAggregator",
    "BetaAggregator",
    "GammaAggregator",
    "FEGaussianAggregator",
    "REGaussianAggregator",
    "HistogramAggregator",
]


class SingletonAggregator(ExperimentAggregator):
    """An aggregation to apply to an ExperimentGroup that needs no aggregation.

    For example, the ExperimentGroup only contains one Experiment.

    Essentially just the [identity function](https://en.wikipedia.org/wiki/Identity_function):

    $$f(x)=x$$

    """

    full_name = "Singleton experiment aggregation"
    aliases = ["singleton", "identity"]

    def aggregate(  # noqa: D102
        self,
        experiment_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
        bounds: tuple[float, float],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        num_samples, num_experiments = experiment_samples.shape

        if num_experiments > 1:
            raise ValueError(
                f"Parameter `num_experiments` > 1. Currently {num_experiments}",
            )

        return experiment_samples[:, 0]


class BetaAggregator(ExperimentAggregator):
    r"""Samples from the beta-conflated distribution.

    Specifically, the aggregate distribution $\text{Beta}(\tilde{\alpha}, \tilde{\beta})$ is
    estimated as:

    $$\begin{aligned}
        \tilde{\alpha}&=\left[\sum_{i=1}^{M}\alpha_{i}\right]-\left(M-1\right) \
        \tilde{\beta}&=\left[\sum_{i=1}^{M}\beta_{i}\right]-\left(M-1\right)
    \end{aligned}$$

    where $M$ is the total number of experiments.

    Uses [`scipy.stats.beta`](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.beta.html)
    class to fit beta-distributions.

    Danger: Assumptions:
        - the individual experiment distributions are beta distributed
        - the metrics are bounded, although the range need not be (0, 1)

    References: Read more:
        1. [Hill, T. P. (2008). Conflations Of Probability Distributions: An Optimal Method For
            Consolidating Data From Different Experiments.](http://arxiv.org/abs/0808.1808)
        2. [Hill, T. P., & Miller, J. (2011). How to combine independent data sets for the same quantity.](https://arxiv.org/abs/1005.4978)
        3. ['Beta distribution' on Wikipedia](https://en.wikipedia.org/wiki/Beta_distribution)

    Args:
        estimation_method (str): method for estimating the parameters of the individual
            experiment distributions. Options are 'mle' for maximum-likelihood estimation, or 'mome'
            for the method of moments estimator. MLE tends be more efficient but is difficult
            to estimate

    """

    full_name = "Beta conflated experiment aggregation"
    aliases = ["beta", "beta_conflation"]

    def __init__(self, rng: RNG, *, estimation_method: str = "mle") -> None:
        super().__init__(rng=rng)

        # Honestly should get rid of this
        # MLE is more efficient than MoME, and difference is small
        self.estimation_method = estimation_method

    def aggregate(  # noqa: D102
        self,
        experiment_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
        bounds: tuple[float, float],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        num_samples, num_experiments = experiment_samples.shape

        if bounds[0] == -float("inf") or bounds[1] == float("inf"):
            raise NotImplementedError(
                "Beta aggregation does not (yet) support metrics with infite bounds.",
            )

        alphas = []
        betas = []
        for per_experiment_samples in experiment_samples.T:
            alpha, beta, _, _ = scipy.stats.beta.fit(
                np.clip(
                    per_experiment_samples,
                    a_min=bounds[0] + 1e-9,
                    a_max=bounds[1] - 1e-9,
                ),
                method=self.estimation_method,
                floc=bounds[0],
                fscale=bounds[1] - bounds[0],
            )

            alphas.append(alpha)
            betas.append(beta)

        conflated_alpha = sum(alphas) - (num_experiments - 1)
        conflated_beta = sum(betas) - (num_experiments - 1)

        # Scipy distributions won't accept the RNG wrapper
        # So pass rng.rng
        conflated_distribution_samples: jtyping.Float[np.ndarray, " num_samples"] = (
            scipy.stats.beta.rvs(
                a=conflated_alpha,
                b=conflated_beta,
                size=num_samples,
                loc=bounds[0],
                scale=bounds[1] - bounds[0],
                random_state=self.rng.rng,
            )
        )  # type: ignore

        # conflated_distribution_samples = (
        #    bounds[1] - bounds[0]
        # ) * conflated_distribution_samples + bounds[0]

        return conflated_distribution_samples


class GammaAggregator(ExperimentAggregator):
    r"""Samples from the Gamma-conflated distribution.

    Specifically, the aggregate distribution $\text{Gamma}(\tilde{\alpha}, \tilde{\beta})$
    ($\alpha$ is the shape, $\beta$ the rate parameter) is estimated as:

    $$\begin{aligned}
        \tilde{\alpha}&=\left[\sum_{i}^{M}\alpha_{i}\right]-(M-1) \
        \tilde{\beta}&=\dfrac{1}{\sum_{i}^{M}\beta_{i}^{-1}}
    \end{aligned}$$

    where $M$ is the total number of experiments.

    An optional `shifted: bool` argument exists to dynamically estimate the support for the
    distribution. Can help fit to individual experiments, but likely minimally impacts the
    aggregate distribution.

    Danger: Assumptions:
        - the individual experiment distributions are gamma distributed

    References: Read more:
        1. [Hill, T. (2008). Conflations Of Probability Distributions: An Optimal Method For Consolidating Data From Different Experiments.](http://arxiv.org/abs/0808.1808)
        2. [Hill, T., & Miller, J. (2011). How to combine independent data sets for the same quantity.](https://arxiv.org/abs/1005.4978)
        3. ['Gamma distribution' on Wikipedia](https://en.wikipedia.org/wiki/Gamma_distribution)

    """  # noqa: E501

    full_name = "Gamma conflated experiment aggregator"
    aliases = ["gamma", "gamma_conflation"]

    def __init__(self, rng: RNG, *, shifted: bool = False) -> None:
        super().__init__(rng=rng)

        self.shifted = shifted

    def aggregate(  # noqa: D102
        self,
        experiment_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
        bounds: tuple[float, float],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        if bounds[0] < 0 or bounds[1] < 0:
            raise ValueError(
                "Gamma aggregator does not support metrics with negative bounds.",
            )
        num_samples, num_experiments = experiment_samples.shape

        # Estimate the 'loc' variable, i.e. the minimum of the support
        # Improves fit to individual experiment distributions
        # Minimal impact on the conflated distribution
        loc_estimate = np.min(experiment_samples) - 1e-09 if self.shifted else bounds[0]

        # Estimate the shape and rate for each distribution
        alphas = []
        betas = []
        for per_experiment_samples in experiment_samples.T:
            finite_samples = per_experiment_samples[np.isfinite(per_experiment_samples)]

            if finite_samples.shape[0] < 0.9 * per_experiment_samples.shape[0]:
                warnings.warn(
                    "An experiment sample has more than 10% non-finite values.",
                )

            alpha, _, beta = scipy.stats.gamma.fit(
                finite_samples,
                floc=loc_estimate,
            )

            alphas.append(alpha)
            betas.append(beta)

        alphas = np.array(alphas)
        betas = np.array(betas)

        # Estimate the parameters of the conflated distribution
        conflated_alpha = np.sum(alphas) - (num_experiments - 1)
        conflated_beta = 1 / np.sum(1 / betas)

        # Redefine the sampling distribution
        conflated_distribution = scipy.stats.gamma(
            a=conflated_alpha,
            scale=conflated_beta,
            loc=loc_estimate,
        )

        # Sample from the distribution, truncating at the bounds of the metric
        conflated_distribution_samples = truncated_sample(
            sampling_distribution=conflated_distribution,
            bounds=(loc_estimate, bounds[1]),
            rng=self.rng,
            num_samples=num_samples,
        )

        return conflated_distribution_samples


class FEGaussianAggregator(ExperimentAggregator):
    r"""Samples from the Gaussian-conflated distribution.

    This is equivalent to the fixed-effects meta-analytical estimator.

    Uses the inverse variance weighted mean and standard errors. Specifically, the aggregate
    distribution $\mathcal{N}(\tilde{\mu}, \tilde{\sigma})$ is estimated as:

    $$\begin{aligned}
        w_{i}&=\dfrac{\sigma_{i}^{-2}}{\sum_{j}^{M}\sigma_{j}^{-2}} \\
        \tilde{\mu}&=\sum_{i}^{M}w_{i}\mu_{i} \\
        \tilde{\sigma}^2&=\dfrac{1}{\sum_{i}^{M}\sigma_{i}^{-2}}
    \end{aligned}$$

    where $M$ is the total number of experiments.

    Danger: Assumptions:
        - the individual experiment distributions are normally (Gaussian) distributed
        - there **is no** inter-experiment heterogeneity present

    References: Read more:
        1. [Hill, T. (2008). Conflations Of Probability Distributions: An Optimal Method For Consolidating Data From Different Experiments.](http://arxiv.org/abs/0808.1808)
        2. [Hill, T., & Miller, J. (2011). How to combine independent data sets for the same quantity.](https://arxiv.org/abs/1005.4978)
        3. [Higgins, J., & Thomas, J. (Eds.). (2023). Cochrane handbook for systematic reviews of interventions.](https://training.cochrane.org/handbook/current/chapter-10#section-10-3)
        4. [Borenstein et al. (2021). Introduction to meta-analysis.](https://www.wiley.com/en-us/Introduction+to+Meta-Analysis%2C+2nd+Edition-p-9781119558354)
        5. ['Meta-analysis' on Wikipedia](https://en.wikipedia.org/wiki/Meta-analysis#Statistical_models_for_aggregate_data)

    """  # noqa: E501

    full_name = "Fixed-effect Gaussian meta-analytical experiment aggregator"
    aliases = ["fe", "fixed_effect", "fe_gaussian", "gaussian", "normal", "fe_normal"]

    def aggregate(  # noqa: D102
        self,
        experiment_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
        bounds: tuple[float, float],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        num_samples, num_experiments = experiment_samples.shape

        # Estimate the means and variances for each distribution
        means = np.mean(experiment_samples, axis=0)
        variances = np.var(experiment_samples, axis=0, ddof=1)

        # Compute the aggregated mean and variance
        # i.e. the inverse-variance weighted mean
        weights = 1 / variances

        agg_mu = np.sum(weights * means) / np.sum(weights)
        agg_var = 1 / np.sum(weights)

        # Redefine the sampling distribution
        conflated_distribution = scipy.stats.norm(
            loc=agg_mu,
            scale=np.sqrt(agg_var),
        )

        # Sample from the distribution, truncating at the bounds of the metric
        conflated_distribution_samples = truncated_sample(
            sampling_distribution=conflated_distribution,
            bounds=bounds,
            rng=self.rng,
            num_samples=num_samples,
        )

        return conflated_distribution_samples


class REGaussianAggregator(ExperimentAggregator):
    r"""Samples from the Random Effects Meta-Analytical Estimator.

    First uses the standard the inverse variance weighted mean and standard errors as model
    parameters, before debiasing the weights to incorporate inter-experiment heterogeneity.
    As a result, studies with larger standard errors will be upweighted relative to the
    fixed-effects model.

    Specifically, starting with a Fixed-Effects model
    $\mathcal{N}(\tilde{\mu_{\text{FE}}}, \tilde{\sigma_{\text{FE}}})$,

    $$\begin{aligned}
        w_{i}&=\dfrac{\left(\sigma_{i}^2+\tau^2\right)^{-1}}{\sum_{j}^{M}\left(\sigma_{j}^2+\tau^2\right)^{-1}} \\
        \tilde{\mu}&=\sum_{i}^{M}w_{i}\mu_{i} \\
        \tilde{\sigma^2}&=\dfrac{1}{\sum_{i}^{M}\sigma_{i}^{-2}}
    \end{aligned}$$

    where $\tau$ is the estimated inter-experiment heterogeneity, and $M$ is the total number
    of experiments.

    Uses the Paule-Mandel iterative heterogeneity estimator, which does not make a parametric
    assumption. The more common (but biased) DerSimonian-Laird estimator can also be used by setting
    `paule_mandel_heterogeneity: bool = False`.

    If `hksj_sampling_distribution: bool = True`, the aggregated distribution is a more conservative
    $t$-distribution, with degrees of freedom equal to $M-1$. This is especially more conservative
    when there are only a few experiments available, and can substantially increase the aggregated
    distribution's variance.

    Danger: Assumptions:
        - the individual experiment distributions are normally (Gaussian) distributed
        - there **is** inter-experiment heterogeneity present

    References: Read more:
        3. [Higgins, J., & Thomas, J. (Eds.). (2023). Cochrane handbook for systematic reviews of interventions.](https://training.cochrane.org/handbook/current/chapter-10#section-10-3)
        4. [Borenstein et al. (2021). Introduction to meta-analysis.](https://www.wiley.com/en-us/Introduction+to+Meta-Analysis%2C+2nd+Edition-p-9781119558354)
        5. ['Meta-analysis' on Wikipedia](https://en.wikipedia.org/wiki/Meta-analysis#Statistical_models_for_aggregate_data)
        4. [IntHout, J., Ioannidis, J. P., & Borm, G. F. (2014). The Hartung-Knapp-Sidik-Jonkman method for random effects meta-analysis is straightforward and considerably outperforms the standard DerSimonian-Laird method.](https://bmcmedresmethodol.biomedcentral.com/articles/10.1186/1471-2288-14-25)
        5. [Langan et al. (2019). A comparison of heterogeneity variance estimators in simulated random‐effects meta‐analyses.](https://onlinelibrary.wiley.com/doi/full/10.1002/jrsm.1316?casa_token=NcK51p09KsYAAAAA%3A_ZkOpRymLWcDTOK5uv6UCJah6MLuEZ430pJJAENiRq2HF9_K4AlGQqhJ7_akJUig5DxkoiKec1Hdp60)

    Args:
        paule_mandel_heterogeneity (bool): whether to use the Paule-Mandel method for estimating
            inter-experiment heterogeneity, or fallback to the DerSimonian-Laird estimator.
            Defaults to True.
        hksj_sampling_distribution (bool): whether to use the Hartung-Knapp-Sidik-Jonkman corrected
            $t$-distribition as the aggregate sampling distribution.
            Defaults to False.
    """  # noqa: E501

    full_name = "Random-effects Gaussian meta-analytical experiment aggregator"
    aliases = ["re", "random_effect", "re_gaussian", "re_normal"]

    def __init__(
        self,
        rng: RNG,
        *,
        paule_mandel_heterogeneity: bool = True,
        hksj_sampling_distribution: bool = False,
    ) -> None:
        super().__init__(rng=rng)
        self.paule_mandel_heterogeneity = paule_mandel_heterogeneity
        self.hksj_sampling_distribution = hksj_sampling_distribution

    def aggregate(  # noqa: D102
        self,
        experiment_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
        bounds: tuple[float, float],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        num_samples, num_experiments = experiment_samples.shape

        # Estimate the means and variances for each distribution
        means = np.mean(experiment_samples, axis=0)
        variances = np.var(experiment_samples, axis=0, ddof=1)

        # Estimate the between-experiment variance
        tau2 = heterogeneity_dl(means, variances)
        if self.paule_mandel_heterogeneity:
            tau2 = heterogeneity_pm(
                means,
                variances,
                init_tau2=tau2,
                maxiter=100,
                # This is still *very* new, experimental
                # Leave False for now
                use_viechtbauer_correction=False,
            )

        weights = 1 / (variances + tau2)

        agg_variance = 1 / np.sum(weights)
        agg_mean = np.sum(weights * means) / np.sum(weights)

        if self.hksj_sampling_distribution:
            # Uses t-distrbution instead of normal distribution
            # More conservative for small number of studies
            q = np.sum(weights * np.power(means - agg_mean, 2)) / (num_experiments - 1)

            # HKSJ factor with correction
            # Strictly more conservative than FE model
            # hksj_factor = np.sqrt(q)
            hksj_factor = max(1.0, np.sqrt(q))

            aggregated_distribution = scipy.stats.t(
                df=num_experiments - 1,
                loc=agg_mean,
                scale=hksj_factor * np.sqrt(agg_variance),
            )

        else:
            # Uses Gaussian distrbution
            aggregated_distribution = scipy.stats.norm(
                loc=agg_mean,
                scale=np.sqrt(agg_variance),
            )

        aggregated_distribution_samples = truncated_sample(
            sampling_distribution=aggregated_distribution,
            bounds=bounds,
            rng=self.rng,
            num_samples=num_samples,
        )

        return aggregated_distribution_samples


class HistogramAggregator(ExperimentAggregator):
    r"""Samples from a histogram approximate conflation distribution.

    First bins all individual experiment groups, and then computes the product of the probability
    masses across individual experiments.

    Unlike other methods, this does not make a parametric assumption. However, the resulting
    distribution can 'look' unnatural, and requires overlapping supports within the sample.
    If any experiment assigns 0 probability mass to any bin, the conflated bin will also
    contain 0 probability mass.

    As such, inter-experiment heterogeneity can be a significant problem.

    Uses [numpy.histogram_bin_edges](https://numpy.org/doc/stable/reference/generated/numpy.histogram_bin_edges.html)
    to estimate the number of bin edges needed per experiment, and takes the smallest across all
    experiments for the aggregate distribution.

    Danger: Assumptions:
        - the individual experiment distributions' supports overlap

    References: Read more:
        1. [Hill, T. (2008). Conflations Of Probability Distributions: An Optimal Method For Consolidating Data From Different Experiments.](http://arxiv.org/abs/0808.1808)
        2. [Hill, T., & Miller, J. (2011). How to combine independent data sets for the same quantity.](https://arxiv.org/abs/1005.4978)

    """  # noqa: E501

    full_name = "Histrogram approximated conflation experiment aggregation"
    aliases = ["hist", "histogram"]

    def __init__(
        self,
        rng: RNG,
        pseudo_count_weight: float = 0.1,
    ) -> None:
        super().__init__(rng=rng)

        # This is super arbitrary and should probably be tuned
        self.pseudo_count_weight = pseudo_count_weight

    def aggregate(  # noqa: D102
        self,
        experiment_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
        bounds: tuple[float, float],
    ) -> jtyping.Float[np.ndarray, " num_samples"]:
        num_samples, num_experiments = experiment_samples.shape

        # Find the smallest recommended bin width for all experiments
        min_bin_width = float("inf")
        for per_experiment_samples in experiment_samples.T:
            distribution_bins = np.histogram_bin_edges(
                per_experiment_samples,
                bins="auto",
            )

            bin_width = distribution_bins[2] - distribution_bins[1]

            if bin_width < min_bin_width:
                min_bin_width = bin_width

        # Find the support for the aggregated histogram
        # Avoids having lots of zero-count bins
        min_min = np.min(experiment_samples)
        max_max = np.max(experiment_samples)

        found_bins = np.arange(
            start=max(min_min - min_bin_width, bounds[0]),
            stop=min(max_max + 2 * min_bin_width, bounds[1]),
            step=min_bin_width,
        )
        num_bins = found_bins.shape[0]

        # The pseudo-counts should have `pseudo_count_weight` times the weight of the true samples
        smoothing_coeff = 1 / num_bins * self.pseudo_count_weight * num_samples

        conflated_distribution = np.zeros(shape=(num_bins - 1,))
        for per_experiment_samples in experiment_samples.T:
            # Re-compute the histograms along the support of the aggregated distribution
            binned_distribution, bins = np.histogram(
                per_experiment_samples,
                bins=found_bins,
                range=bounds,
            )

            # Estimate the bin probabilities
            log_p_hat = np.log(binned_distribution + smoothing_coeff) - np.log(
                num_samples + smoothing_coeff * num_bins,
            )

            conflated_distribution += log_p_hat

        conflated_distribution = np.exp(
            conflated_distribution - np.logaddexp.reduce(conflated_distribution),
        )

        # Resample the conflated distribution
        # Samples at the midpoint of each bin
        conflated_distribution_samples = self.rng.choice(
            (bins[:-1] + bins[1:]) / 2,  # type: ignore
            size=num_samples,
            p=conflated_distribution,
        )

        # Jitter the values so they fall off the bin midpoints
        bin_noise = self.rng.uniform(
            low=-min_bin_width / 2,
            high=min_bin_width / 2,
            size=num_samples,
        )

        conflated_distribution_samples = np.clip(
            conflated_distribution_samples + bin_noise,
            a_min=bounds[0],
            a_max=bounds[1],
        )

        return conflated_distribution_samples



import pathlib
import numpy as np
import pytest

from prob_conf_mat.io import (
    validate_confusion_matrix,
    load_csv,
    ConfMatIOWarning,
    ConfMatIOError,
)

DATA_DIR = pathlib.Path("./tests/data")

assert DATA_DIR.exists() and DATA_DIR.is_dir(), DATA_DIR

class TestCSV:
    def test_file_not_exist(self):
        with pytest.raises(FileNotFoundError, match="No such file or directory:"):
            load_csv(
                location="foobarbaz.csv",
            )

    @pytest.mark.parametrize(
        argnames="file_path",
        argvalues=list((DATA_DIR / "confusion_matrices").glob("*.csv")),
    )
    def test_valid_csv(self, file_path):
        load_csv(
            location=file_path,
        )

    @pytest.mark.parametrize(
        argnames="file_path",
        argvalues=list((DATA_DIR / "malformed_confusion_matrices").glob("*.csv")),
    )
    def test_malformed_csv(self, file_path):
        with pytest.raises(
            expected_exception=ConfMatIOError,
        ):
            load_csv(
                location=file_path,
            )


class TestConfMatValidation:
    def test_dtype_conversion(self) -> None:
        # Integer should pass directly
        validate_confusion_matrix(confusion_matrix=[[1, 0], [0, 1]])

        # Object should fail
        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is not of type integer.",
        ):
            validate_confusion_matrix(confusion_matrix=[[1, "foo"], [0, 1]])

        # Float should fail
        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is not of type integer.",
        ):
            validate_confusion_matrix(confusion_matrix=[[1.0, 0], [0, 1]])

        # uint should not fail
        validate_confusion_matrix(
            confusion_matrix=np.array([[1, 0], [0, 1]], dtype=np.uint),
        )

        # complex float should fail
        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is not of type integer.",
        ):
            validate_confusion_matrix(
                confusion_matrix=np.array([[1, 0], [0, 1]], dtype=np.complex128),
            )

        # bool should not fail
        validate_confusion_matrix(
            confusion_matrix=np.array([[1, 0], [0, 1]], dtype=np.bool),
        )

        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is not of type integer.",
        ):
            validate_confusion_matrix(confusion_matrix=np.array([[1, 0], [np.inf, 1]]))

    def test_shape(self) -> None:
        # 2D Square matrix should not fail
        conf_mat = np.ones((2, 2), dtype=np.int64)
        validate_confusion_matrix(confusion_matrix=conf_mat)

        # Non 2D matrix should fail
        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is malformed.",
        ):
            conf_mat = np.ones((3, 3, 3), dtype=np.int64)
            validate_confusion_matrix(confusion_matrix=conf_mat)

        # Non square matrix should fail
        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is malformed.",
        ):
            conf_mat = np.ones((2, 4), dtype=np.int64)
            validate_confusion_matrix(confusion_matrix=conf_mat)

        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is malformed.",
        ):
            conf_mat = np.ones((4, 2), dtype=np.int64)
            validate_confusion_matrix(confusion_matrix=conf_mat)

        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix is malformed.",
        ):
            conf_mat = np.ones((1, 1), dtype=np.int64)
            validate_confusion_matrix(confusion_matrix=conf_mat)

    def test_empty(self) -> None:
        with pytest.raises(ConfMatIOError, match="Some rows contain no entries"):
            validate_confusion_matrix(confusion_matrix=[[0, 0], [0, 1]])

        with pytest.warns(
            ConfMatIOWarning,
            match="Some columns contain no entries, meaning model never predicted it.",
        ):
            validate_confusion_matrix(confusion_matrix=[[0, 1], [0, 1]])

    def test_value_bounds(self) -> None:
        with pytest.raises(
            ConfMatIOError,
            #match="The loaded confusion matrix has non-finite elements.",
        ):
            validate_confusion_matrix(confusion_matrix=[[1, np.nan], [0, 1]])

        with pytest.raises(
            ConfMatIOError,
            match="The loaded confusion matrix has negative elements.",
        ):
            validate_confusion_matrix(confusion_matrix=[[1, -1], [0, 1]])



from .summary import summarize_posterior, PosteriorSummary
from .batched_averaging import (
    numpy_batched_arithmetic_mean,
    numpy_batched_convex_combination,
    numpy_batched_geometric_mean,
    numpy_batched_harmonic_mean,
)
from .dirichlet_distribution import (
    _DIRICHLET_PRIOR_STRATEGIES,
    dirichlet_prior,
    dirichlet_sample,
)
from .score_interval import wilson_score_interval
from .truncated_sampling import truncated_sample
from .hdi_estimation import hdi_estimator



import jaxtyping as jtyping
import numpy as np
import pytest
import scipy.stats

from prob_conf_mat.utils.rng import RNG
from prob_conf_mat.experiment_aggregation import get_experiment_aggregator
from prob_conf_mat.stats.truncated_sampling import truncated_sample

SEED = 0
DIMS = 8
NUM_EXPERIMENTS = 1
NUM_SAMPLES = 10000

NP_RNG = np.random.default_rng(seed=SEED)

NON_SINGLETON_AGG_METHODS = [
    "beta",
    "fe_gaussian",
    "gamma",
    "histogram",
    "re_gaussian",
]


def unbounded_samples(
    rng,
    num_experiments: int,
    num_samples: int,
) -> jtyping.Float[
    np.ndarray,
    "num_samples num_experiments dims",
]:
    samples = rng.normal(
        loc=0.5,
        scale=0.1,
        size=(num_samples, num_experiments),
    )

    return samples


def bounded_samples(
    rng,
    num_experiments: int,
    num_samples: int,
    bounds: tuple[float, float],
) -> jtyping.Float[
    np.ndarray,
    "num_samples num_experiments dims",
]:
    samples = np.stack(
        arrays=[
            truncated_sample(
                sampling_distribution=scipy.stats.norm(
                    loc=0.5,
                    scale=0.1,
                ),
                bounds=bounds,
                rng=rng,
                num_samples=num_samples,
            )
            for _ in range(num_experiments)
        ],
        axis=1,
    )

    return samples


class TestSingletonAggregator:
    aggregator = get_experiment_aggregator(
        aggregation="singleton",
        rng=RNG(seed=SEED),
    )

    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            unbounded_samples(rng=NP_RNG, num_experiments=1, num_samples=10000),
        ],
    )
    def test_valid(
        self,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        self.aggregator.aggregate(
            experiment_samples=samples,
            bounds=(-float("inf"), float("inf")),
        )

    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            unbounded_samples(rng=NP_RNG, num_experiments=2, num_samples=10000),
        ],
    )
    def test_too_many_experiments(
        self,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        with pytest.raises(
            expected_exception=ValueError,
            match="Parameter `num_experiments` > 1.",
        ):
            self.aggregator.aggregate(
                experiment_samples=samples,
                bounds=(-float("inf"), float("inf")),
            )

    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            unbounded_samples(rng=NP_RNG, num_experiments=1, num_samples=10000),
        ],
    )
    def test_values(
        self,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        output_samples = self.aggregator.aggregate(
            experiment_samples=samples,
            bounds=(-float("inf"), float("inf")),
        )

        np.allclose(samples, output_samples)


class TestNonSingletonAggregator:
    @pytest.mark.parametrize("aggregation", NON_SINGLETON_AGG_METHODS)
    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            bounded_samples(
                rng=NP_RNG,
                num_experiments=2,
                num_samples=10000,
                bounds=(0, 1),
            ),
        ],
    )
    def test_valid(
        self,
        aggregation: str,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        aggregator = get_experiment_aggregator(
            aggregation=aggregation,
            rng=RNG(seed=SEED),
        )

        aggregator.aggregate(
            experiment_samples=samples,
            bounds=(0, 1),
        )

    @pytest.mark.parametrize("aggregation", NON_SINGLETON_AGG_METHODS)
    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            unbounded_samples(
                rng=NP_RNG,
                num_experiments=2,
                num_samples=10000,
            ),
        ],
    )
    def test_infinite_bounds(
        self,
        aggregation: str,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        aggregator = get_experiment_aggregator(
            aggregation=aggregation,
            rng=RNG(seed=SEED),
        )

        match aggregation:
            case "beta":
                with pytest.raises(NotImplementedError):
                    aggregator.aggregate(
                        experiment_samples=samples,
                        bounds=(-float("inf"), float("inf")),
                    )

            case "gamma":
                with pytest.raises(ValueError):
                    aggregator.aggregate(
                        experiment_samples=samples,
                        bounds=(-float("inf"), float("inf")),
                    )

            case _:
                aggregator.aggregate(
                    experiment_samples=samples,
                    bounds=(-float("inf"), float("inf")),
                )

    @pytest.mark.parametrize("aggregation", NON_SINGLETON_AGG_METHODS)
    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            bounded_samples(
                rng=NP_RNG,
                num_experiments=2,
                num_samples=10000,
                bounds=(0, 1),
            ),
        ],
    )
    def test_bounds(
        self,
        aggregation: str,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        aggregator = get_experiment_aggregator(
            aggregation=aggregation,
            rng=RNG(seed=SEED),
        )

        out_samples = aggregator.aggregate(
            experiment_samples=samples,
            bounds=(0, 1),
        )

        assert np.max(out_samples) <= 1.0
        assert np.min(out_samples) >= 0.0

    @pytest.mark.parametrize("aggregation", NON_SINGLETON_AGG_METHODS)
    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            bounded_samples(
                rng=NP_RNG,
                num_experiments=2,
                num_samples=10000,
                bounds=(0.4, 0.75),
            ),
        ],
    )
    def test_uneven_bounds(
        self,
        aggregation: str,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        aggregator = get_experiment_aggregator(
            aggregation=aggregation,
            rng=RNG(seed=SEED),
        )

        out_samples = aggregator.aggregate(
            experiment_samples=samples,
            bounds=(0.40, 0.75),
        )

        assert np.max(out_samples) <= 0.75
        assert np.min(out_samples) >= 0.40

    @pytest.mark.parametrize("aggregation", NON_SINGLETON_AGG_METHODS)
    @pytest.mark.parametrize(
        argnames="samples",
        argvalues=[
            bounded_samples(
                rng=NP_RNG,
                num_experiments=2,
                num_samples=10000,
                bounds=(0, 1),
            ),
        ],
    )
    def test_variance(
        self,
        aggregation: str,
        samples: jtyping.Float[np.ndarray, "num_samples num_experiments"],
    ) -> None:
        aggregator = get_experiment_aggregator(
            aggregation=aggregation,
            rng=RNG(seed=SEED),
        )

        out_samples = aggregator.aggregate(
            experiment_samples=samples,
            bounds=(0, 1),
        )

        assert np.all(np.var(out_samples) <= np.var(samples, axis=0)), (
            np.var(out_samples),
            np.var(samples, axis=0),
        )



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

    from prob_conf_mat.utils import MetricLike

from dataclasses import dataclass

import numpy as np


@dataclass(frozen=True)
class ListwiseComparisonResult:
    """Class to store results from a listwise comparison between experiments."""

    experiment_names: list[str]
    metric: MetricLike

    p_experiment_given_rank: jtyping.Float[
        np.ndarray,
        " num_experiments num_experiments",
    ]
    p_rank_given_experiment: jtyping.Float[
        np.ndarray,
        " num_experiments num_experiments",
    ]


def listwise_compare(
    experiment_scores_dict: dict[str, jtyping.Float[np.ndarray, " num_samples"]],
    metric: MetricLike,
) -> ListwiseComparisonResult:
    """Compares all experiments against each other in a listwise manner.

    This essentially tries to estimate to the expected ranking of the experiments.

    Args:
        experiment_scores_dict (dict[str, Float[ndarray, "num_samples" ]]): a mapping of experiment
            name to sampled metric values
        metric (MetricLike): the metric used. This parameter is only used for reporting

    Returns:
        ListwiseComparisonResult: the comparison result
    """
    # Convert the experiment values to a list of lists
    all_experiment_scores = list(experiment_scores_dict.items())

    # Transpose the list of lists
    experiment_names, experiment_scores = list(
        map(list, zip(*all_experiment_scores)),
    )

    # Stack the experiments into a [num_samples, num_experiments] array
    experiment_scores = np.stack(
        arrays=experiment_scores,
        axis=1,
    )

    num_samples, num_experiments = experiment_scores.shape

    # TODO: pre-sort the arrays by the mean scores
    # *Might* speed up quicksort
    # Very low priority
    # Rank the arrays
    experiment_inv_ranks = np.argsort(np.argsort(experiment_scores, axis=1), axis=1)

    # Invert the ranking (largest value gets smallest rank)
    experiment_ranks = experiment_scores.shape[1] - experiment_inv_ranks

    # Count the number of times each experiment achieved a certain rank
    rank_count_matrix = np.zeros((num_experiments, num_experiments))
    for i in range(num_experiments):
        ranks, counts = np.unique(experiment_ranks[:, i], axis=0, return_counts=True)

        for rank, rank_count in zip(ranks, counts):
            rank_count_matrix[i, rank - 1] = rank_count

    # Normalize the matrix
    rank_prob_matrix = rank_count_matrix / num_samples

    # Sort the table by MRR
    reciprocal_rank = 1 / (np.arange(rank_prob_matrix.shape[0]) + 1)

    mrr = np.sum(
        reciprocal_rank[np.newaxis, :] * rank_prob_matrix,
        axis=1,
    )

    idx = np.argsort(1 - mrr)
    rank_prob_matrix = rank_prob_matrix[idx, :]

    result = ListwiseComparisonResult(
        experiment_names=[experiment_names[i] for i in idx],
        metric=metric,
        p_rank_given_experiment=rank_prob_matrix,
        p_experiment_given_rank=rank_prob_matrix.T,
    )

    return result



from __future__ import annotations


import typing

from prob_conf_mat.experiment_comparison import pairwise_compare

if typing.TYPE_CHECKING:
    from prob_conf_mat.metrics import Metric, AveragedMetric


def pairwise_comparison(
    study,
    metric: Metric | AveragedMetric,
    class_label: int,
    experiment_group_a: str,
    experiment_group_b: str,
    experiment_a: str | typing.Literal["aggregated"],
    experiment_b: str | typing.Literal["aggregated"],
    min_sig_diff: typing.Optional[float] = None,
    precision: int = 4,
) -> None:
    lhs_samples = study.get_metric_samples(
        metric=metric.name,
        experiment_group=experiment_group_a,
        experiment=experiment_a,
        sampling_method="posterior",
    ).values[:, class_label]

    rhs_samples = study.get_metric_samples(
        metric=metric.name,
        experiment_group=experiment_group_b,
        experiment=experiment_b,
        sampling_method="posterior",
    ).values[:, class_label]

    lhs_random_samples = study.get_metric_samples(
        metric=metric.name,
        experiment_group=experiment_group_a,
        experiment=experiment_a,
        sampling_method="random",
    ).values[:, class_label]

    rhs_random_samples = study.get_metric_samples(
        metric=metric.name,
        experiment_group=experiment_group_b,
        experiment=experiment_b,
        sampling_method="random",
    ).values[:, class_label]

    if experiment_a != "aggregated" and experiment_b != "aggregated":
        lhs_observed = study.get_metric_samples(
            metric=metric.name,
            experiment_group=experiment_group_a,
            experiment=experiment_a,
            sampling_method="input",
        ).values[:, class_label]

        rhs_observed = study.get_metric_samples(
            metric=metric.name,
            experiment_group=experiment_group_b,
            experiment=experiment_b,
            sampling_method="input",
        ).values[:, class_label]

        observed_diff = lhs_observed - rhs_observed
    else:
        observed_diff = None

    comparison_result = pairwise_compare(
        metric=metric,
        diff_dist=lhs_samples - rhs_samples,
        random_diff_dist=lhs_random_samples - rhs_random_samples,
        ci_probability=study.ci_probability,
        min_sig_diff=min_sig_diff,
        observed_difference=observed_diff,
        lhs_name=f"{experiment_group_a}/{experiment_a}",
        rhs_name=f"{experiment_group_b}/{experiment_b}",
    )

    print(comparison_result.template_sentence(precision=precision))



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import numpy as np
    import jaxtyping as jtyping

    from prob_conf_mat.metrics import MetricCollection
    from prob_conf_mat.experiment import ExperimentResult
    from prob_conf_mat.experiment_aggregation import ExperimentAggregator
    from prob_conf_mat.experiment_aggregation.abc import ExperimentAggregationResult
    from prob_conf_mat.utils import RNG, MetricLike

from collections import OrderedDict
from warnings import warn
from dataclasses import dataclass

from prob_conf_mat.experiment import Experiment, ExperimentResult
from prob_conf_mat.experiment_aggregation import get_experiment_aggregator


@dataclass(frozen=True)
class ExperimentGroupResult:
    """A wrapper class for the output of an ExperimentGroup."""

    aggregation_result: OrderedDict[MetricLike, ExperimentAggregationResult]
    individual_experiment_results: OrderedDict[
        MetricLike,
        dict[Experiment, ExperimentResult],
    ]


class ExperimentGroup:
    """This class represents a group of related Experiments.

    The results across experiments can be aggregated to give an average result for the group.

    For example, this could represent the same model evaluated across different folds of the same
    dataset. Or they could be results on te same dataset from models with different
    weight initializations.

    Args:
        name (str): the name of this experiment group
        rng (RNG): the RNG used to control randomness
    """

    def __init__(
        self,
        name: str,
        rng: RNG,
    ) -> None:
        self.name = name

        # ======================================================================
        # Import hyperparameters
        # ======================================================================
        # The manager's RNG
        self.rng = rng

        # The collection of experiments
        self.num_classes = None
        self.experiments: OrderedDict = OrderedDict()

    @property
    def num_experiments(self) -> int:
        """The number of experiments in this ExperimentGroup."""
        return len(self.experiments)

    def __len__(self) -> int:
        """The number of experiments in this ExperimentGroup."""
        return self.num_experiments

    def add_experiment(
        self,
        name: str,
        confusion_matrix: jtyping.Int[np.ndarray, " num_classes num_classes"],
        prevalence_prior: str
        | float
        | jtyping.Float[np.typing.ArrayLike, " num_classes"] = 0,
        confusion_prior: str
        | float
        | jtyping.Float[np.typing.ArrayLike, " num_classes num_classes"] = 0,
    ) -> None:
        """Adds an Experiment to this ExperimentGroup.

        Each experiment is characterized by a single confusion matrix.

        Args:
            name (str): the name of this experiment
            confusion_matrix (jtyping.Int[np.ndarray, 'num_classes num_classes']): the
                confusion matrix for this experiment.
            prevalence_prior (str | float | Float[ArrayLike, ' num_classes'], optional): the prior
                over the prevalence counts for this experiments.
                Defaults to 0, Haldane's prior.
            confusion_prior (str | float | Float[ArrayLike, ' num_classes num_classes'], optional):
                the prior over the confusion counts for this experiments.
                Defaults to 0, Haldane's prior.

        """
        # Spawn a new RNG, that is independent of the RNGs used in other experiments of this
        # experiment group
        indep_rng = self.rng.spawn(1)[0]

        # Define the experiment
        new_experiment = Experiment(
            # Provided for the user
            # Unique to each experiment
            name=name,
            confusion_matrix=confusion_matrix,
            prevalence_prior=prevalence_prior,
            confusion_prior=confusion_prior,
            # Provided by the manager
            rng=indep_rng,
        )

        # Check to make sure that the number of classes in the new experiment matches the number of
        # classes in all other experiment
        # This really is the minimal requirement for an experiment group
        if self.num_classes is None:
            self.num_classes = new_experiment.num_classes
        elif new_experiment.num_classes != self.num_classes:
            raise AttributeError(
                (
                    f"Experiment '{self.name}/{name}' has {new_experiment.num_classes} classes, "
                    f"not the expected {self.num_classes}"
                ),
            )

        # Check if this experiment already exists
        # Overwrite if so
        if self.experiments.get(name, None) is not None:  # type: ignore
            warn(
                message=f"Experiment '{self.name}/{name}' already exists. Overwriting.",
            )

        # Finally, add the experiment to the experiment store
        self.experiments[name] = new_experiment

    def __getitem__(self, key: str) -> Experiment:
        """Gets an experiment under this ExperimentGroup by its name.

        Args:
            key (str): the experiment name

        Returns:
            Experiment: _description_
        """
        if key not in self.experiments:
            raise KeyError(
                (
                    f"No experiment with name '{key}' is currently present in experiment group "
                    f"'{self.name}'."
                ),
            )

        return self.experiments[key]

    def sample_metrics(  # noqa: D102
        self,
        metrics: MetricCollection,
        sampling_method: str,
        num_samples: int,
        metric_to_aggregator: dict[MetricLike, ExperimentAggregator],
    ) -> ExperimentGroupResult:
        """Samples and aggregates metrics from Experiments belonging to this ExperimentGroup.

        Args:
            metrics (MetricCollection): the collection of metrics for which samples are needed.
            sampling_method (str): the sampling method used to construct
                synthetic confusion matrices.
            num_samples (int): the number of synthetic confusion matrices to sample.
            metric_to_aggregator (_type_): how each metric should be aggregated.

        Returns:
            ExperimentGroupResult: the output of all Experiments and their aggregation
        """
        # Compute metrics for each experiment and store them
        all_metrics_experiment_results: dict[
            MetricLike,
            list[ExperimentResult],
        ] = {metric: [] for metric in metrics.get_insert_order()}
        for _, experiment in self.experiments.items():
            all_metrics_experiment_result: dict[MetricLike, ExperimentResult] = (
                experiment.sample_metrics(
                    metrics=metrics,
                    sampling_method=sampling_method,
                    num_samples=num_samples,
                )
            )

            for metric, experiment_result in all_metrics_experiment_result.items():
                all_metrics_experiment_results[metric].append(experiment_result)

        # Iterate over all the individual experiments and aggregate them
        all_metrics_experiment_aggregation_result = dict()
        for metric, experiment_results in all_metrics_experiment_results.items():
            # Fetch the experiment aggregation method
            aggregator: ExperimentAggregator | None = metric_to_aggregator.get(
                metric,
                None,
            )

            # Handle missing experiment aggregation methods
            if aggregator is None:
                if len(self) > 1:
                    raise ValueError(
                        (
                            f"Metric '{metric.name}' does not have an assigned aggregation method, "
                            f"but experiment group '{self.name}' has {len(self)} experiments. "
                            f"Try adding one using "
                            f"`Study.add_metric(metric={metric.name}, aggregation=...)`."
                        ),
                    )

                # We're allowed to pass the group's RNG, because the singleton aggregator is just
                # an identity function
                aggregator = get_experiment_aggregator(
                    aggregation="singleton",
                    rng=self.rng,
                )

            # Run the aggregation
            experiment_aggregation_result: dict[
                MetricLike,
                ExperimentAggregationResult,
            ] = aggregator(
                experiment_group=self,
                metric=metric,
                experiment_results=experiment_results,
            )  # type: ignore

            all_metrics_experiment_aggregation_result[metric] = (
                experiment_aggregation_result
            )

        # Clean the output
        # All metrics in insertion order
        # Have a nested dict for the experiment results
        all_metrics_experiment_results_cleaned: OrderedDict[
            MetricLike,
            dict[Experiment, ExperimentResult],
        ] = OrderedDict(
            [
                (
                    metric,
                    {
                        experiment_result.experiment: experiment_result
                        for experiment_result in all_metrics_experiment_results[metric]
                    },
                )
                for metric in metrics.get_insert_order()
            ],
        )

        all_metrics_experiment_aggregation_result_cleaned: OrderedDict[
            MetricLike,
            ExperimentAggregationResult,
        ] = OrderedDict(
            [
                (metric, all_metrics_experiment_aggregation_result[metric])
                for metric in metrics.get_insert_order()
            ],
        )

        return ExperimentGroupResult(
            aggregation_result=all_metrics_experiment_aggregation_result_cleaned,
            individual_experiment_results=all_metrics_experiment_results_cleaned,
        )

    def __repr__(self) -> str:  # noqa: D105
        return f"ExperimentGroup({self.name})"

    def __str__(self) -> str:  # noqa: D105
        return f"ExperimentGroup({self.name})"



# Import all required packages
# Flattens the importtime profile
# These take up ~92% of cold import time
import numpy
import scipy
import scipy.stats

# Import the actual Study object
from .study import Study



from pathlib import Path

import pytest
import matplotlib.pyplot as plt

import prob_conf_mat as pcm
from prob_conf_mat.io import load_csv

BASIC_METRICS = [
    "acc",
    "f1",
    "f1@macro",
]

FORMATS = [
    "simple",  # tabulate
    "github",  # tabulate
    "html",  # tabulate
    "records",  # records
    "pandas",  # pandas
    "pd",  # pandas
]

CONF_MAT_PATHS = Path(
    "./tests/data/mnist_digits",
).resolve()


@pytest.fixture(scope="module")
def study() -> pcm.Study:
    study = pcm.Study(
        seed=0,
        num_samples=10000,
        ci_probability=0.95,
    )

    # Add a bucnh of metrics
    for metric in BASIC_METRICS:
        study.add_metric(metric=metric, aggregation="fe_gaussian")

    # Add a bunch of experiments
    for file_path in sorted(CONF_MAT_PATHS.glob("*.csv")):
        # Split the file name to recover the model and fold
        model, fold = file_path.stem.split("_")

        # Load in the confusion matrix using the utility function
        confusion_matrix = load_csv(location=file_path)

        # Add the experiment to the study
        study.add_experiment(
            experiment_name=f"{model}/fold_{fold}",
            confusion_matrix=confusion_matrix,
            prevalence_prior=0,
            confusion_prior=0,
        )

    return study


class TestReportingMethods:
    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_plot_metric_summaries(self, study, metric):
        study.plot_metric_summaries(
            metric=metric,
            class_label=0,
            method="kde",
        )

        plt.close()

        study.plot_metric_summaries(
            metric=metric,
            class_label=0,
            method="hist",
        )

        plt.close()

    @pytest.mark.parametrize(argnames="format", argvalues=FORMATS)
    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_report_random_metric_summaries(self, study, metric, format):
        study.report_random_metric_summaries(
            metric=metric,
            class_label=0,
            table_fmt=format,
        )

    @pytest.mark.parametrize(argnames="format", argvalues=FORMATS)
    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_report_aggregated_metric_summaries(self, study, metric, format):
        study.report_aggregated_metric_summaries(
            metric=metric,
            class_label=0,
            table_fmt=format,
        )

    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_plot_experiment_aggregation(self, study, metric):
        study.plot_experiment_aggregation(
            metric=metric,
            class_label=0,
            experiment_group="mlp",
        )

        study.plot_experiment_aggregation(
            metric=metric,
            class_label=0,
            experiment_group="svm",
        )

        plt.close()

    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_forest_plot(self, study, metric):
        study.plot_forest_plot(metric=metric, class_label=0)

        plt.close()

    @pytest.mark.parametrize(argnames="format", argvalues=FORMATS)
    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_pairwise_comparison(self, study, metric, format):
        study.report_pairwise_comparison(
            metric=metric,
            class_label=0,
            experiment_a="mlp/aggregated",
            experiment_b="svm/aggregated",
            min_sig_diff=0.005,
        )

        with pytest.warns(match="does not produce a table"):
            study.report_pairwise_comparison(
                metric=metric,
                class_label=0,
                experiment_a="mlp/aggregated",
                experiment_b="svm/aggregated",
                min_sig_diff=0.005,
                table_fmt=format,
            )

    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_pairwise_comparison_plot(self, study, metric):
        study.plot_pairwise_comparison(
            metric=metric,
            class_label=0,
            method="kde",
            experiment_a="mlp/aggregated",
            experiment_b="svm/aggregated",
            min_sig_diff=0.005,
        )

        study.plot_pairwise_comparison(
            metric=metric,
            class_label=0,
            method="histogram",
            experiment_a="mlp/aggregated",
            experiment_b="svm/aggregated",
            min_sig_diff=0.005,
        )

        study.plot_pairwise_comparison(
            metric=metric,
            class_label=0,
            method="kde",
            experiment_a="mlp/fold_0",
            experiment_b="svm/fold_0",
            min_sig_diff=0.005,
        )

        plt.close()

    @pytest.mark.parametrize(argnames="format", argvalues=FORMATS)
    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_report_pairwise_comparison_to_random(self, study, metric, format):
        study.report_pairwise_comparison_to_random(
            metric=metric,
            class_label=0,
            table_fmt=format,
        )

    @pytest.mark.parametrize(argnames="format", argvalues=FORMATS)
    @pytest.mark.parametrize(argnames="metric", argvalues=BASIC_METRICS)
    def test_report_listwise_comparison(self, study, metric, format):
        study.report_listwise_comparison(
            metric=metric,
            class_label=0,
            table_fmt=format,
        )



from __future__ import annotations

import typing

from prob_conf_mat.utils import fmt
from prob_conf_mat.experiment_comparison import pairwise_compare

if typing.TYPE_CHECKING:
    from prob_conf_mat.metrics import Metric, AveragedMetric


def pairwise_comparison_to_random(
    study,
    metric: Metric | AveragedMetric,
    class_label: int,
    min_sig_diff: typing.Optional[float] = None,
    precision: int = 4,
    tablefmt: str = "html",
    **tabulate_kwargs,
) -> str:
    # Import optional dependencies only now
    import tabulate

    table = []
    for experiment_group_name, experiment_group in study.experiment_groups.items():
        for experiment_name, experiment in experiment_group.experiments.items():
            sampled_values = study.get_metric_samples(
                metric=metric.name,
                experiment_group=experiment_group_name,
                experiment=experiment_name,
                sampling_method="posterior",
            ).values[:, class_label]

            sampled_random_values = study.get_metric_samples(
                metric=metric.name,
                experiment_group=experiment_group_name,
                experiment=experiment_name,
                sampling_method="random",
            ).values[:, class_label]

            result = pairwise_compare(
                metric=metric,
                diff_dist=sampled_values - sampled_random_values,
                random_diff_dist=None,
                ci_probability=study.ci_probability,
                min_sig_diff=min_sig_diff,
                lhs_name=f"{experiment_group.name}/{experiment.name}",
                rhs_name="random",
                observed_difference=None,
            )

            rope = [-result.min_sig_diff, result.min_sig_diff]

            if rope[1] - rope[0] > 1e-4:
                rope_str = f"[{fmt(rope[0], precision=precision, mode='f')}, {fmt(rope[1], precision=precision, mode='f')}]"
            else:
                rope_str = f"[{fmt(rope[0], precision=precision, mode='e')}, {fmt(rope[1], precision=precision, mode='e')}]"

            row = [
                experiment_group.name,
                experiment.name,
                result.diff_dist_summary.median,
                result.p_direction,
                rope_str,
                result.p_rope,
                result.p_bi_sig,
                result.p_sig_pos,
                result.p_sig_neg,
            ]

            table.append(row)

    headers = [
        "Group",
        "Experiment",
        "Median Δ",
        "$p_{dir}$",
        "ROPE",
        "$p_{ROPE}$",
        "$p_{sig}$",
        "$p_{sig}^+$",
        "$p_{sig}^-$",
    ]

    summary_table = tabulate.tabulate(
        tabular_data=table,
        headers=headers,
        floatfmt=f".{precision}f",
        colalign=["left", "left"] + ["decimal" for _ in headers[2:]],
        tablefmt=tablefmt,
        **tabulate_kwargs,
    )

    return summary_table



from __future__ import annotations
import typing

import pytest

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

import numpy as np
import scipy.stats as stats
import itertools

from prob_conf_mat.utils import RNG

ALL_SEEDS = np.random.default_rng(seed=0).integers(low=0, high=2**31 - 1, size=(10,))


def generate_tree(meta_seed: int, root_seed: int, max_children: int) -> list[RNG]:
    meta_rng = np.random.default_rng(meta_seed)

    tree = [RNG(seed=root_seed)]

    i = 0
    while (i < len(tree) - 1) or (i == 0):
        node = tree[i]

        if len(node.position) == 0:
            child_nodes = node.spawn(
                n_children=int(meta_rng.integers(low=2, high=max_children)),
            )

            tree.extend(child_nodes)

        elif len(node.position) < 4:
            child_nodes = node.spawn(
                n_children=int(meta_rng.integers(low=0, high=max_children)),
            )

            tree.extend(child_nodes)

        i += 1

    return tree


def p_independent(
    node_a: RNG,
    node_b: RNG,
    sample_size: int = 10000,
    num_bins: int = 10,
) -> float:
    node_a_samples = node_a.random(size=(sample_size,))
    node_b_samples = node_b.random(size=(sample_size,))

    bin_counts, _, _ = np.histogram2d(
        x=node_a_samples,
        y=node_b_samples,
        bins=num_bins,
        range=((0, 1), (0, 1)),
    )

    p = stats.power_divergence(f_obs=bin_counts, axis=None, lambda_=0).pvalue  # type: ignore

    return p


def num_rejected_hypotheses(
    p_vals: jtyping.Float[np.ndarray, " num_tests"],
    alpha: float,
) -> int:
    """Uses Holm–Bonferroni's step-down procedure to count the number of rejected null-hypotheses.

    This should correct for the family-wise error rate.

    Args:
        p_vals (jtyping.Float[np.ndarray, "num_tests"]): the p values of all conducted tests
        alpha (float): the significance level

    Returns:
        int: the number of hypotheses that should be rejected according to the provided significance level
    """  # noqa: E501
    fail_to_reject_null = np.sort(p_vals) > alpha / (
        p_vals.shape[0] - np.arange(p_vals.shape[0])
    )

    n_rejected = (
        np.where(fail_to_reject_null)[0][0] if np.any(fail_to_reject_null) else 0
    )

    return n_rejected


# Tests ========================================================================
class TestRNG:
    @pytest.mark.parametrize(argnames="seed", argvalues=ALL_SEEDS)
    def test_equivalence_to_numpy(self, seed: int):
        # Test root node
        our_rng = RNG(seed=seed)
        np_rng = np.random.default_rng(seed)

        our_value = our_rng.random(size=(1000,))
        np_value = np_rng.random(size=(1000,))

        assert np.allclose(our_value, np_value)

        # Test child node
        our_child_rng = our_rng.spawn(1)[0]
        np_child_rng = np.random.default_rng(seed).spawn(1)[0]

        our_value = our_child_rng.random(size=(1000,))
        np_value = np_child_rng.random(size=(1000,))

        assert np.allclose(our_value, np_value)

    @pytest.mark.parametrize(argnames="seed", argvalues=ALL_SEEDS)
    def test_independence(self, seed: int):
        tree = generate_tree(meta_seed=seed, root_seed=0, max_children=4)

        all_pvals = []
        for node_a, node_b in itertools.combinations(tree, 2):
            all_pvals.append(
                p_independent(node_a, node_b, sample_size=10000, num_bins=100),
            )

        all_pvals = np.stack(all_pvals)

        n_rejected = num_rejected_hypotheses(p_vals=all_pvals, alpha=0.1)

        assert n_rejected == 0, n_rejected

    @pytest.mark.parametrize(
        argnames="seed_a,seed_b",
        argvalues=itertools.combinations(ALL_SEEDS, 2),
    )
    def test_changing_seed(self, seed_a: int, seed_b: int):
        tree_generator_seed = abs(seed_a - seed_b)

        # Sample a tree
        tree_a = generate_tree(
            meta_seed=tree_generator_seed,
            root_seed=seed_a,
            max_children=4,
        )

        samples_a = []
        for node in tree_a:
            samples_a.append(node.random())

        samples_a = np.stack(samples_a)

        # Sample a second, independent tree
        tree_b = generate_tree(
            meta_seed=tree_generator_seed,
            root_seed=seed_b,
            max_children=4,
        )

        samples_b = []
        for node in tree_b:
            samples_b.append(node.random())

        samples_b = np.stack(samples_b)

        #! Samples from different trees with different seeds should be independent
        assert not np.allclose(samples_a, samples_b)

        # Set the seed on the root node to the same seed as tree_b
        tree_a[0].seed = seed_b

        samples_a_reset = []
        for node in tree_a:
            samples_a_reset.append(node.random())

        samples_a_reset = np.stack(samples_a_reset)

        #! Samples from different trees with the same seed should be the same
        #! Setting the seed on the root node should reset the tree's state
        assert np.allclose(samples_a_reset, samples_b)

        # Set the seed back to the initial seed of tree_a
        tree_a[0].seed = seed_a

        samples_a_reset = []
        for node in tree_a:
            samples_a_reset.append(node.random())

        samples_a_reset = np.stack(samples_a_reset)

        #! Samples from different trees with the same seed should be the same
        #! Setting the seed on the root node should reset the tree's state
        assert np.allclose(samples_a_reset, samples_a)



from __future__ import annotations

import typing
from collections import deque, OrderedDict
from graphlib import TopologicalSorter
from functools import cache

if typing.TYPE_CHECKING:
    from prob_conf_mat.utils.typing import MetricLike

from prob_conf_mat.metrics.interface import get_metric
from prob_conf_mat.metrics.abc import RootMetric, Metric, AveragedMetric


@cache
def generate_metric_computation_schedule(
    metrics: typing.Sequence[MetricLike],
) -> typing.Sequence[MetricLike]:
    """Generates a topological ordering of the inserted metrics and their dependencies.

    Ensures no function is computed before its dependencies are available.

    Args:
        metrics (Iterable[str | Type[Metric] | Type[AggregatedMetric]]): a iterable of metrics

    Returns:
        typing.Generator[Type[Metric] | Type[AggregatedMetric]]
    """

    seen_metrics = set()
    stack = deque(metrics)
    topological_sorter = TopologicalSorter()
    while len(stack) > 0:
        metric = stack.popleft()

        if isinstance(metric, str):
            metric = get_metric(metric)

        if metric in seen_metrics:
            continue

        for dependency in metric.dependencies:
            dependency = get_metric(dependency)

            topological_sorter.add(metric, dependency)

            stack.append(dependency)

        seen_metrics.add(metric)

    computation_schedule = topological_sorter.static_order()

    return tuple(computation_schedule)


class MetricCollection:
    """Endows a list of metrics with various needed properties, like:
        - metric syntax interfacing
        - redundancy checking
        - topological sorting

    Args:
        metrics (typing.Collection[str | MetricLike | typing.Iterable[str | MetricLike] | typing.Self], optional): the initial collection of metrics. Defaults to ().
    """

    def __init__(
        self,
        metrics: typing.Optional[
            str | MetricLike | typing.Iterable[str | MetricLike] | typing.Self
        ] = (),
    ) -> None:
        self._metrics = OrderedDict()
        self._metrics_by_alias_or_name = dict()

        if metrics is not None:
            self.add(metrics)

    def add(
        self,
        metric: str
        | MetricLike
        | typing.Iterable[str | MetricLike]
        | typing.Self,
    ) -> None:
        """Adds a metric to the metric collection. The 'metric' must be one of:
            - a valid metric syntax string
            - an instance of `Metric` or `AveragedMetric`
            - an iterable of the above two
            - a `MetricCollection`

        Args:
            metric (typing.Collection[str | MetricLike | typing.Iterable[str | MetricLike] | typing.Self], optional): the metric to be added
        """

        # If metric is a str or MetricLike
        if (
            isinstance(metric, str)
            or issubclass(metric.__class__, Metric)
            or issubclass(metric.__class__, AveragedMetric)
            or issubclass(metric.__class__, RootMetric)
        ):
            self._add_metric(metric=metric) # type: ignore

        # If metric is a list of MetricLikes
        elif (
            isinstance(metric, list)
            or isinstance(metric, set)
            or isinstance(metric, tuple)
        ):
            for m in metric:
                self.add(metric=m)

        # If metric is a MetricCollection
        elif isinstance(metric, self.__class__):
            for m in metric.get_insert_order():
                self.add(metric=m)

        # Otherwise
        else:
            raise ValueError(
                f"Cannot process input of type `{type(metric)}` into a MetricCollection."
            )

    def _add_metric(self, metric: str | MetricLike) -> None:
        # Convert str to a MetricLike
        if isinstance(metric, str):
            metric_instance = get_metric(metric)
            self._metrics_by_alias_or_name.update({metric: metric_instance})

        # Simply store a MetricLike
        elif (
            issubclass(metric.__class__, Metric)
            or issubclass(metric.__class__, AveragedMetric)
            or issubclass(metric.__class__, RootMetric)
        ):
            metric_instance = metric

        # Otherwise
        else:
            raise TypeError(
                f"Metric must be of type `str`, or a subclass of `Metric` or `AggregatedMetric`, not {metric}: {type(metric)}"  # noqa: E501
            )

        self._metrics.update(((metric_instance, None),))
        self._metrics_by_alias_or_name.update({metric_instance.name: metric_instance})
        #self._metrics_by_alias_or_name.update(
        #    {alias: metric_instance for alias in metric_instance.aliases}
        #)

    def get_insert_order(self) -> tuple[MetricLike]:
        return tuple(self._metrics.keys())

    def get_compute_order(self) -> "MetricCollection":
        topologically_sorted = generate_metric_computation_schedule(
            self.get_insert_order()
        )

        return MetricCollection(metrics=topologically_sorted)

    def __getitem__(self, key: str | MetricLike):
        return self._metrics_by_alias_or_name[key]

    def __iter__(self) -> typing.Generator[MetricLike]:
        for metric in self.get_insert_order():
            yield metric

    def __len__(self):
        return len(self._metrics)

    def __repr__(self) -> str:
        return f"MetricCollection({list(self._metrics.keys())})"



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from typing import Literal

    import jaxtyping as jtyping

from dataclasses import dataclass

import numpy as np
import scipy

from prob_conf_mat.utils.formatting import fmt


def heterogeneity_dl(
    means: jtyping.Float[np.ndarray, " num_experiments"],
    variances: jtyping.Float[np.ndarray, " num_experiments"],
) -> float:
    """Compute the DerSimonian-Laird estimate of between-experiment heterogeneity.

    Args:
        means (Float[np.ndarray, "num_experiments"]): the experiment means
        variances (Float[np.ndarray, "num_experiments"]): the experiment variances

    Returns:
        float: estimate of the between-experiment heterogeneity
    """
    num_experiments = means.shape[0]

    w_fe = 1 / variances

    mu_fe = np.sum(w_fe * means) / np.sum(w_fe)

    q = np.sum(w_fe * np.power(means - mu_fe, 2))
    df = num_experiments - 1
    denom = np.sum(w_fe) - (np.sum(np.power(w_fe, 2)) / np.sum(w_fe))
    tau2 = max(0.0, (q - df) / denom)

    return tau2


def heterogeneity_pm(
    means: jtyping.Float[np.ndarray, " num_experiments"],
    variances: jtyping.Float[np.ndarray, " num_experiments"],
    init_tau2: float = 0.0,
    atol: float = 1e-5,
    maxiter: int = 100,
    *,
    use_viechtbauer_correction: bool = False,
) -> float:
    """Compute the Paule-Mandel estimate of between-experiment heterogeneity.

    Based on the `_fit_tau_iterative` function from `stats_models`.

    Original implementation is based on Appendix A of [1]

    We make two modifications:
        1. instead of stopping iteration if F(tau_2) < 0, we back-off to the midpoint
            between the current and previous estimate
        2. optionally, we apply the Viechtbauer correction to the root. Instead of
            converging to the mean, converge to the median

    References: Read More:
        1. DerSimonian, R., & Kacker, R. (2007). Random-effects model for meta-analysis
            of clinical trials: an update. Contemporary clinical trials, 28(2), 105-114.


    Args:
        means (Float[np.ndarray, " num_experiments"]): the experiment means
        variances (Float[np.ndarray, " num_experiments"]): the experiment variances
        init_tau2 (float, optional): the inital tau2 estimate. Defaults to 0.0.
        atol (float, optional): when to assume convergence. Defaults to 1e-5.
        maxiter (int, optional): the maximum number of iterations needed.
            Defaults to 50.
        use_viechtbauer_correction (bool, optional): whether to use the Viechtbauer
            correction. Very new. Defaults to False.

    Returns:
        float: estimate of the between-experiment heterogeneity
    """
    prev_tau2 = 0.0
    tau2 = init_tau2
    num_experiments = means.shape[0]
    patience = maxiter

    if use_viechtbauer_correction:
        root = scipy.stats.chi2(df=num_experiments - 1).median()
    else:
        root = num_experiments - 1

    while patience > 0:
        # Estimate RE summary stat
        w_pm = 1 / (variances + tau2)
        mu_pm = np.sum(w_pm * means) / np.sum(w_pm)

        # Compute generalised Q values
        # i.e. the residual squares
        q_pm = np.sum(w_pm * np.power(means - mu_pm, 2))

        # Check if residual has converged
        f_tau2 = q_pm - root
        if f_tau2 < 0:
            # If negative, reduce tau2 estimate
            # and try again
            # This is different from other implementations
            tau2 = (prev_tau2 + tau2) / 2

            # Allow for an extra iteration for the correction to take effect
            patience -= 1
            continue
            # break

        if np.allclose(f_tau2, 0.0, atol=atol):
            # Check convergence
            break

        # Otherwise, update tau2
        delta_denom = np.sum(np.power(w_pm, 2) * np.power(means - mu_pm, 2))

        prev_tau2 = tau2
        tau2 += f_tau2 / delta_denom

        patience -= 1

    if tau2 < 0.0:
        tau2 = 0

    return tau2


@dataclass(frozen=True)
class HeterogeneityResult:
    """The output of a heterogeneity computation."""

    i2: float
    within_experiment_variance: float
    between_experiment_variance: float
    i2_interpretation: str

    def template_sentence(self, precision: int = 4) -> str:
        """Fills a template string with some standard summary statistics."""
        within_experiment_variance = fmt(
            self.within_experiment_variance,
            precision=precision,
            mode="f",
        )

        between_experiment_variance = fmt(
            self.between_experiment_variance,
            precision=precision,
            mode="f",
        )

        template_sentence = (
            f"I2 is {fmt(self.i2, precision=precision, mode='%')}"
            f" (variance within={within_experiment_variance}, "
            f"between={between_experiment_variance})."
            f"\nThis can be considered '{self.i2_interpretation}'."
        )

        return template_sentence


def estimate_i2(
    individual_samples: jtyping.Float[np.ndarray, " num_samples num_experiments"],
) -> HeterogeneityResult:
    """Estimates a generalised $I^2$ metric, as suggested by Bowden et al. [1].

    It measures the amount of variance attributable to within-experiment variance vs.
    between-experiment variance. The between experiment variance is estimated using a
    Paule-Mandel tau2 estimator.

    References: Read more:
        1. Bowden, J., Tierney, J. F., Copas, A. J., & Burdett, S. (2011). Quantifying,
            displaying and accounting for heterogeneity in the meta-analysis of RCTs
            using standard and generalised Qstatistics. BMC medical research
            methodology, 11(1), 1-12.

    Args:
        individual_samples (Float[np.ndarray, "num_samples num_experiments"]): the
            samples from individual experiments

    Returns:
        float: the $I^2$ estimate
    """
    if individual_samples.shape[1] == 1:
        result = HeterogeneityResult(
            i2=1.0,
            within_experiment_variance=float(np.var(individual_samples)),
            between_experiment_variance=0.0,
            i2_interpretation=interpret_i2(i2_score=0.0),
        )

        return result

    means = np.mean(individual_samples, axis=0)
    variances = np.var(individual_samples, axis=0)

    # Pooled intra-experiment variance
    # Assuming equal population sizes
    pooled_var = np.mean(variances)

    # Estimate of inter-experiment variance
    tau2 = heterogeneity_dl(means, variances)
    tau2 = heterogeneity_pm(
        means,
        variances,
        init_tau2=tau2,
        use_viechtbauer_correction=False,
    )

    # Proportion of variance attributable to inter-experiment variance
    i2 = tau2 / (tau2 + pooled_var)

    result = HeterogeneityResult(
        i2=i2,
        within_experiment_variance=pooled_var,
        between_experiment_variance=tau2,
        i2_interpretation=interpret_i2(i2),
    )

    return result


def interpret_i2(
    i2_score: float,
) -> Literal[
    "insignificant heterogeneity",
    "borderline moderate heterogeneity",
    "moderate heterogeneity",
    "borderline substantial heterogeneity",
    "borderline considerable heterogeneity",
    "considerable heterogeneity",
    "unknown",
]:
    """Interprets $I^2$ values using prescribed guidelines [1].

    References: Read More:
        1. Higgins, J. P., & Green, S. (Eds.). (2008). Cochrane handbook for
            systematic reviews of interventions.

    Args:
        i2_score (float): the I2 estimate

    Returns:
        str: a rough interpretation of the magnitude of I2
    """
    het_sig = ""

    if i2_score < 0.0 or i2_score > 1.0:
        raise ValueError(
            f"I^2 should be in the range (0.0, 1.0). Currently: {i2_score}",
        )
    if i2_score < 0.3:
        het_sig = "insignificant"
    elif i2_score >= 0.3 and i2_score < 0.4:
        het_sig = "borderline moderate"
    elif i2_score >= 0.4 and i2_score < 0.5:
        het_sig = "moderate"
    elif i2_score >= 0.5 and i2_score < 0.6:
        het_sig = "borderline substantial"
    elif i2_score >= 0.6 and i2_score < 0.75:
        het_sig = "borderline considerable"
    elif i2_score >= 0.75 and i2_score <= 1.0:
        het_sig = "considerable"
    else:
        # Edge case for NaN results
        # i.e., division by 0
        return "unknown"

    het_sig += " heterogeneity"

    return het_sig



import typing

from prob_conf_mat.metrics import RootMetric, Metric, AveragedMetric

MetricType: typing.TypeVar = typing.TypeVar('MetricType', bound=Metric)
AveragedMetricType: typing.TypeVar = typing.TypeVar('AveragedMetricType', bound=AveragedMetric)
RootMetricType: typing.TypeVar = typing.TypeVar('RootMetricType', bound=RootMetric)

MetricLike: typing.TypeAlias = typing.Union[RootMetric, Metric, AveragedMetric]



from __future__ import annotations
import typing

from prob_conf_mat.stats import summarize_posterior
from prob_conf_mat.utils import fmt

if typing.TYPE_CHECKING:
    from prob_conf_mat.metrics import Metric, AveragedMetric
    from tabulate import tabulate


def experiment_summaries(
    study,
    metric: Metric | AveragedMetric,
    class_label: int,
    precision: int = 4,
    table_fmt: str = "html",
    **tabulate_kwargs,
) -> str:
    # Import optional dependencies only now
    import tabulate

    table = []
    for experiment_group_name, experiment_group in study.experiment_groups.items():
        for experiment_name, experiment in experiment_group.experiments.items():
            observed_experiment_result = study.get_metric_samples(
                metric=metric.name,
                experiment_group=experiment_group_name,
                experiment=experiment_name,
                sampling_method="input",
            )

            sampled_experiment_result = study.get_metric_samples(
                metric=metric.name,
                experiment_group=experiment_group_name,
                experiment=experiment_name,
                sampling_method="posterior",
            )

            distribution_summary = summarize_posterior(
                sampled_experiment_result.values[:, class_label],
                ci_probability=study.ci_probability,
            )

            if distribution_summary.hdi[1] - distribution_summary.hdi[0] > 1e-4:
                hdi_str = f"[{fmt(distribution_summary.hdi[0], precision=precision, mode='f')}, {fmt(distribution_summary.hdi[1], precision=precision, mode='f')}]"
            else:
                hdi_str = f"[{fmt(distribution_summary.hdi[0], precision=precision, mode='e')}, {fmt(distribution_summary.hdi[1], precision=precision, mode='e')}]"

            table_row = [
                experiment_group_name,
                experiment_name,
                observed_experiment_result.values[:, class_label],
                distribution_summary.median,
                distribution_summary.mode,
                hdi_str,
                distribution_summary.metric_uncertainty,
                distribution_summary.skew,
                distribution_summary.kurtosis,
            ]

            table.append(table_row)

    headers = ["Group", "Experiment", "Observed", *distribution_summary.headers]  # type: ignore

    table = tabulate.tabulate(  # type: ignore
        tabular_data=table,
        headers=headers,
        floatfmt=f".{precision}f",
        colalign=["left", "left"] + ["decimal" for _ in headers[2:]],
        tablefmt=table_fmt,
        **tabulate_kwargs,
    )

    return table


def random_experiment_summaries(
    study,
    metric: Metric | AveragedMetric,
    class_label: int,
    precision: int = 4,
    table_fmt: str = "html",
    **tabulate_kwargs,
):
    table = []
    for experiment_group_name, experiment_group in study.experiment_groups.items():
        for experiment_name, experiment in experiment_group.experiments.items():
            sampled_experiment_result = study.get_metric_samples(
                metric=metric.name,
                experiment_group=experiment_group_name,
                experiment=experiment_name,
                sampling_method="random",
            )

            distribution_summary = summarize_posterior(
                sampled_experiment_result.values[:, class_label],
                ci_probability=study.ci_probability,
            )

            if distribution_summary.hdi[1] - distribution_summary.hdi[0] > 1e-4:
                hdi_str = f"[{fmt(distribution_summary.hdi[0], precision=precision, mode='f')}, {fmt(distribution_summary.hdi[1], precision=precision, mode='f')}]"
            else:
                hdi_str = f"[{fmt(distribution_summary.hdi[0], precision=precision, mode='e')}, {fmt(distribution_summary.hdi[1], precision=precision, mode='e')}]"

            table_row = [
                experiment_group_name,
                experiment_name,
                distribution_summary.median,
                distribution_summary.mode,
                hdi_str,
                distribution_summary.metric_uncertainty,
                distribution_summary.skew,
                distribution_summary.kurtosis,
            ]

            table.append(table_row)

    headers = ["Group", "Experiment", *distribution_summary.headers]  # type: ignore

    table = tabulate.tabulate(  # type: ignore
        tabular_data=table,
        headers=headers,
        floatfmt=f".{precision}f",
        colalign=["left", "left"] + ["decimal" for _ in headers[2:]],
        tablefmt=table_fmt,
        **tabulate_kwargs,
    )

    return table



def changes_state(method):
    """Whenever the state of a study changes, make sure to clean the cache"""

    def inner(self, *args, **kwargs):
        self.cache.clean()
        return method(self, *args, **kwargs)

    return inner



import numpy as np

from prob_conf_mat import Study
from prob_conf_mat.metrics import get_metric
from prob_conf_mat.metrics.abc import Metric, Averaging
from prob_conf_mat.experiment_aggregation import get_experiment_aggregator
from prob_conf_mat.experiment_aggregation.abc import ExperimentAggregator
from prob_conf_mat.utils.rng import RNG


class TestExtension:
    def test_defining_new_metric(self) -> None:
        # First define the metric
        class FowlkesMallows(Metric):
            full_name = "Fowlkes Mallows Index"
            is_multiclass = False
            bounds = (0.0, 1.0)
            dependencies = ("ppv", "tpr")
            sklearn_equivalent = "fowlkes_mallows_index"
            aliases = ["fowlkes_mallows", "fm", "fmi"]

            def compute_metric(self, ppv, tpr):
                return np.sqrt(ppv * tpr)

        # Then make sure it is accessible through the metric interface
        get_metric("fowlkes_mallows")

        # Finally, make sure a standard study example works
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_experiment(
            "test/test",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=1,
            confusion_prior=1,
        )

        study.add_metric("fowlkes_mallows")

        study.report_metric_summaries("fowlkes_mallows", class_label=0)

    def test_defining_new_averaging_method(self) -> None:
        # First define the averaging method
        class Take2ndClass(Averaging):
            """Always takes the value of the 2nd class.

            ...
            """

            full_name = "Takes 2nd Class Value"
            dependencies = ()
            sklearn_equivalent = "binary, with positive_class=1"
            aliases = ["2nd_class", "two"]

            def compute_average(self, metric_values):
                scalar_array = metric_values[:, 1]

                return scalar_array

        # Then make sure it is accessible through the interface
        get_metric("f1@two")

        # Finally, make sure a standard study example works
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_experiment(
            "test/test",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=1,
            confusion_prior=1,
        )

        study.add_metric("f1@two")

        study.report_metric_summaries("f1@two", class_label=0)

    def test_defining_new_composed_metric(self) -> None:
        # Make sure it is accessible through the metric interface
        get_metric("fowlkes_mallows@two")

        # Finally, make sure a standard study example works
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_experiment(
            "test/test",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=1,
            confusion_prior=1,
        )

        study.add_metric("fowlkes_mallows@two")

        study.report_metric_summaries("fowlkes_mallows@two", class_label=0)

    def test_defining_new_experiment_aggregator(self) -> None:
        # First define the experiment aggregator
        class Take1stExperiment(ExperimentAggregator):
            full_name: str = "Always Takes 1st Experiment Result as Aggregate"
            aliases: list[str] = ["first", "1st"]

            def aggregate(
                self,
                experiment_samples,
                bounds,
            ):
                return experiment_samples[:, 1]

        # Then make sure it is accessible through the interface
        get_experiment_aggregator(aggregation="first", rng=RNG(0))

        # Finally, make sure a standard study example works
        study = Study(seed=0, num_samples=10000, ci_probability=0.95)

        study.add_experiment(
            "test/test_a",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=1,
            confusion_prior=1,
        )

        study.add_experiment(
            "test/test_b",
            confusion_matrix=[[1, 0], [0, 1]],
            prevalence_prior=1,
            confusion_prior=1,
        )

        study.add_metric("f1", aggregation="first")

        study.report_aggregated_metric_summaries("f1", class_label=0)



import dataclasses

import numpy as np
import jaxtyping as jtyping
import scipy


@dataclasses.dataclass(frozen=True, kw_only=True)
class KDEResult:
    """_summary_."""

    kernel: scipy.stats._kde.gaussian_kde
    x_vals: jtyping.Float[np.ndarray, " num_kernel_samples"]
    y_vals: jtyping.Float[np.ndarray, " num_kernel_samples"]
    bounds: tuple[float, float]


def compute_kde(
    samples: jtyping.Float[np.ndarray, " num_samples"],
    bw_method: str | None = None,
    bw_adjust: float = 1,
    cut: float = 3,
    clip: tuple[float, float] | None = None,
    grid_samples: int = 200,
) -> KDEResult:
    """Computes a Gaussian KDE.

    Tries to match the interface used by [Seaborn's KDE estimator](https://github.com/mwaskom/seaborn/blob/master/seaborn/_statistics.py#L42).

    Args:
        samples (Float[ndarray, 'num_samples']): the experiment samples
        bw_method (str | None, optional): the bandwidth selection method to use.
            Defaults to Scipy's default: 'scott'.
        bw_adjust (float, optional):
            Factor that multiplicatively scales the bandwidth chosen.
            Increasing will make the curve smoother.
            Defaults to 1.
        cut (float, optional): Factor, multiplied by the smoothing bandwidth, that determines how
            far the evaluation grid extends past the extreme datapoints.
            When set to 0, truncate the curve at the data limits.
            Defaults to 3.
        clip (tuple[float, float] | None, optional): the bounds outside of which no density values
            will be computed.
            Defaults to None.
        grid_samples (int, optional): the number of KDE points to evaluate at.
            Defaults to 200.

    Returns:
        KDEResult: the output of the KDE
    """
    # Check the space we're allowed to plot in
    empirical_min = np.min(samples)
    empirical_max = np.max(samples)

    min_grid_val = max(
        empirical_min - cut * bw_adjust,
        clip[0] if clip is not None else -np.inf,
    )

    max_grid_val = min(
        empirical_max + cut * bw_adjust,
        clip[1] if clip is not None else np.inf,
    )

    grid_x_vals = np.linspace(min_grid_val, max_grid_val, num=grid_samples)

    # Compute the kernel
    kernel = scipy.stats.gaussian_kde(dataset=samples, bw_method=bw_method)
    kernel.set_bandwidth(bw_method=kernel.factor * bw_adjust)

    # Compute the kernel output
    grid_y_vals = kernel(grid_x_vals)

    result = KDEResult(
        kernel=kernel,
        x_vals=grid_x_vals,
        y_vals=grid_y_vals,
        bounds=(min_grid_val, max_grid_val),
    )

    return result



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    import jaxtyping as jtyping

import pathlib
import warnings
import csv

import numpy as np


class ConfMatIOWarning(Warning):
    """Some warning to highlight potential undesirable behaviour due to IO."""
    ...


class ConfMatIOError(Exception):
    """While trying to perform confusion matrix IO, some exception was encountered."""
    ...

def load_csv(
    location: str | pathlib.Path,
    encoding: str = "utf-8",
    newline: str = "\n",
    dialect: str = "excel",
    delimiter: str = ",",
    lineterminator: str = "\r\n",
    dtype: np.typing.DTypeLike = np.int64,
) -> jtyping.Int[np.ndarray, " num_classes num_classes"]:
    """Loads a CSV file into memory, and parses it as if it were a valid confusion matrix.

    Args:
        location (str | pathlib.Path): the location of csv file containing the confusion matrix
        encoding (str): the encoding of the confusion matrix file
        newline (str): the newline character used in the confusion matrix file
        dialect (str): the csv dialect, passed to `csv.reader`
        delimiter (str): the csv delimiter character, passed to `csv.reader`
        lineterminator (str): the csv lineterminator character, passed to `csv.reader`
        dtype (DTypeLike, optional): the desired dtype of the numpy array. Defaults to int64.

    Returns:
        Int[ndarray, 'num_classes num_classes']: the parsed confusion matrix
    """
    location = pathlib.Path(location)
    rows = []
    with location.open(
        mode="r",
        newline=newline,
        encoding=encoding,  # type: ignore
    ) as f:
        reader = csv.reader(
            f,
            dialect=dialect,
            delimiter=delimiter,
            lineterminator=lineterminator,
        )

        for i, row in enumerate(reader):
            try:
                row_vals = list(map(int, row))
            except ValueError:
                raise ConfMatIOError(
                    f"Row contains values that cannot be converted to int: "
                    f"Row number: {i}. File: {location}",
                )

            rows.append(row_vals)

    try:
        arr = np.array(rows, dtype=dtype)
    except Exception as e:  # noqa: BLE001
        raise ConfMatIOError(
            f"Could not convert loaded csv to a confusion matrix."
            f"Encountered the following exception: {e}",
        )

    return arr


def validate_confusion_matrix(
    confusion_matrix: jtyping.Int[np.typing.ArrayLike, " num_classes num_classes"],
    dtype: np.typing.DTypeLike = np.int64,
) -> jtyping.Int[np.ndarray, " num_classes num_classes"]:
    """Validates a confusion matrix to prevent any future funny business.

    For a confusion matrix to be valid, it:
        1. Must be square matrix (i.e., arrays with 2 dimensions)
        2. Must contain only positive integers
        3. Must contain at least 2 classes
        4. Must have at least one record for each ground-truth class
        5. Should have at least one record for each prediction

    Args:
        confusion_matrix (jtyping.Int[np.ndarray, 'num_classes num_classes']): the confusion matrix
        dtype (DTypeLike, optional): the desired dtype of the numpy array. Defaults to int64.

    Returns:
        Int[ndarray, 'num_classes num_classes']: the validated confusion matrix as a numpy ndarray
    """
    #! Must be an np.ndarray
    if not isinstance(confusion_matrix, np.ndarray):
        try:
            confusion_matrix = np.array(object=confusion_matrix)
        except Exception as e:  # noqa: BLE001
            raise ConfMatIOError(
                f"While trying to convert a confusion matrix to a numpy array, "
                f"encountered the following exception: {e}.",
            )

    #! Must be 2-dimensional
    if not (len(confusion_matrix.shape) == 2):
        raise ConfMatIOError(
            f"The loaded confusion matrix is malformed. "
            f"Shape: {confusion_matrix.shape}. "
            f"A confusion matrix should have exactly 2 dimensions. "
            f"Current dimensions: {confusion_matrix.shape}",
        )

    #! Must be square
    if not (confusion_matrix.shape[0] == confusion_matrix.shape[1]):
        raise ConfMatIOError(
            f"The loaded confusion matrix is malformed. "
            f"Shape: {confusion_matrix.shape}. "
            f"A confusion matrix should be square. "
            f"Current dimensions: {confusion_matrix.shape}",
        )

    #! Must have at least 2 classes
    if confusion_matrix.shape[0] == 1 or confusion_matrix.shape[1] == 1:
        raise ConfMatIOError(
            f"The loaded confusion matrix is malformed. "
            f"Shape: {confusion_matrix.shape}. "
            f"A confusion matrix should have at least 2 classes. "
            f"Current dimensions: {confusion_matrix.shape}",
        )

    #! Must be an array of only integers
    # Or at the very least, a dtype such that `dtype + float = float`
    if not np.issubdtype(confusion_matrix.dtype, np.integer):
        try:
            confusion_matrix = confusion_matrix.astype(dtype=dtype, casting="safe")
        except Exception as e:  # noqa: BLE001
            raise ConfMatIOError(
                f"The loaded confusion matrix is not of type integer. "
                f"While trying to convert, encounterted the following exception: {e}. "
                f"Confusion matrix: {confusion_matrix}",
            )

    #! All values must be finite
    if not np.all(np.isfinite(confusion_matrix)):
        raise ConfMatIOError(
            f"The loaded confusion matrix has non-finite elements. "
            f"Confusion matrix: {confusion_matrix}",
        )

    #! All values must be positive
    if not np.all(confusion_matrix >= 0):
        raise ConfMatIOError(
            f"The loaded confusion matrix has negative elements. "
            f"Confusion matrix: {confusion_matrix}",
        )

    #! Must have at least one record for each ground truth class
    cond_counts = confusion_matrix.sum(axis=1)
    if not np.all(cond_counts > 0):
        offenders = np.where(cond_counts == 0)[0].tolist()
        raise ConfMatIOError(
            f"Some rows contain no entries, meaning condition does not exist. "
            f"Rows: {offenders}. "
            f"Confusion matrix: {confusion_matrix}",
        )

    #! Should have at least one record for each predicted class
    # This is easily violated for poorly optimized models with class imbalance
    pred_counts = confusion_matrix.sum(axis=0)
    if not np.all(pred_counts > 0):
        offenders = np.where(pred_counts == 0)[0].tolist()
        warnings.warn(
            message=(
                f"Some columns contain no entries, meaning model never predicted it. "
                f"Columns: {offenders}. Confusion matrix: {confusion_matrix}"
            ),
            category=ConfMatIOWarning,
        )

    return confusion_matrix


def pred_cond_to_confusion_matrix(
    pred_cond: jtyping.Int[np.ndarray, " num_samples 2"],
    *,
    pred_first: bool = True,
) -> jtyping.Int[np.ndarray, " num_classes num_classes"]:
    """Converts an array-like of model prediction, ground truth pairs into an unnormalized confusion matrix.

    Confusion matrix *always* has predictions on the columns, condition on the rows.

    Args:
        pred_cond (jtyping.Int[np.ndarray, ' num_samples 2']): the arraylike collection of
            predictions
        pred_first (bool, optional): whether the model prediction is on the first column,
            or the ground truth label.
            Defaults to True.

    Returns:
        jtyping.Int[np.ndarray, ' num_classes num_classes']
    """  # noqa: E501
    support = np.unique(pred_cond)
    support_size = support.shape[0]
    if not (np.arange(support_size) == support).all():
        raise ValueError(
            f"Predictions file must contain all labels at least once. "
            f"Found labels for {list(support)}",
        )

    locs, counts = np.unique(pred_cond, axis=0, return_counts=True)

    confusion_matrix = np.zeros((support_size, support_size), dtype=int)

    if pred_first:
        confusion_matrix[locs[:, 1], locs[:, 0]] = counts
    else:
        confusion_matrix[locs[:, 1], locs[:, 0]] = counts

    return confusion_matrix


def confusion_matrix_to_pred_cond(
    confusion_matrix: jtyping.Int[np.ndarray, " num_classes num_classes"],
    *,
    pred_first: bool = True,
) -> jtyping.Int[np.ndarray, " num_samples 2"]:
    """Converts an unnormalized confusion matrix into an array of model prediction,
    ground truth pairs.

    Assumes predictions on the columns, condition on the rows of the confusion matrix.

    Args:
        confusion_matrix (jtyping.Int[np.ndarray, ' num_classes num_classes']): the unnormalized
            confusion matrix
        pred_first (bool, optional): whether the model prediction should be on the first column,
            or the ground truth label.
            Defaults to True.

    Returns:
        jtyping.Int[np.ndarray, ' num_samples 2']
    """  # noqa: D205
    output = []
    for row_num, row in enumerate(confusion_matrix):
        for col_num, occurences in enumerate(row):
            if pred_first:
                output.extend([[col_num, row_num]] * occurences)
            else:
                output.extend([[row_num, col_num]] * occurences)

    output = np.array(output)

    return output
